import subprocess
import unittest

import nzpy


class testnzpy(unittest.TestCase):

    def setUp(self):
        self.setupDb()
        self.conn = nzpy.connect(user="admin", password="password",
                                 host='localhost',
                                 port=5480, database="nzpy_test",
                                 securityLevel=0, logLevel=0)
        self.cursor = self.conn.cursor()

    # common methods for all objects
    @classmethod
    def executeSQL(cls, dbname, query):
        sql = ["nzsql", "-d", dbname, "-Axc", query]
        newProc = subprocess.Popen(sql, stdout=subprocess.PIPE)
        newProc.wait()
        if newProc.returncode != 0:
            raise Exception("Query execution failed:: " + str(dbname) +
                            " : " + str(query))
        return newProc.communicate()[0].decode()

    @classmethod
    def executeShellCmd(cls, cmd):
        newProc = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
        newProc.wait()
        if newProc.returncode != 0:
            raise Exception("Command execution failed:: " + str(cmd))
        print(str(cmd) + " COMPLETED")
        return newProc.communicate()[0].decode()

    @classmethod
    def setupDb(cls):
        try:
            cls.executeSQL("system", "drop database nzpy_test")
        except Exception:
            pass  # consume exception is database does not exist
        try:
            cls.executeSQL("system", "create database nzpy_test")
        except Exception as exp:
            print(exp)

    @classmethod
    def getRowCount(cls, output):
        return int(output.strip().split('|')[1])

    @classmethod
    def performOperation(cls, opName):
        try:
            with open(opName, "r") as fp:
                cmd = fp.readline()
                while cmd:
                    if cmd.split()[0] == "nzsql":
                        dbname = cmd.split()[2]
                        query = cmd.split('"')[1]
                        cls.executeSQL(dbname, query)
                    else:
                        cls.executeShellCmd(cmd.strip())
                    cmd = fp.readline()
        except Exception as exp:
            print(exp)

    def test_Update(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("UPDATE all_datatypes SET a_char20=?, "
                            "a_varchar1=?, "
                            "a_varchar50=?, a_real=?, a_double=?, "
                            "a_byteint=?, a_smallint=?, "
                            "a_int=?, a_bigint=?, a_numps=?, "
                            "a_nump=?, a_date=?, a_time=?, "
                            "a_timetz=?, a_timestamp=?, a_interval=?",
                            ('jeDHTrgMRvsNwftessEg', 'a',
                             'sdfsdrtdsrcsfsdfsd32423fd%$',
                             100, 6.6666, 12, 1111, 200, 111617881,
                             4456456.456367, 56123, '2001-01-01',
                             '12:12:12', '22:22:22+12:21',
                             '1223-12-12 12:12:12',
                             '22 years 22 months 22 days 22 hours 22 minutes '
                             '22 seconds'))
        self.assertEqual(12, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("UPDATE all_datatypes SET a_char20="
                            "'LaN$SCCTn%lWe"
                            "mRvVEBq',a_varchar1='X',a_varchar50="
                            "'aS53d13Dds3c',a_real= "
                            "53.442,a_double=-1234.3456423,a_byteint=-128,"
                            "a_smallint="
                            "32767,a_int=43523565,a_bigint=23145355656334534,"
                            "a_numps="
                            "34524123.456532,a_nump=12345,a_date='1999-09-25',"
                            "a_time='12:37:25',"
                            "a_timetz='20:15:35+5:30',a_timestamp = "
                            "'1999-09-25 12:37:25',"
                            "a_interval='15 years 13 months 25 days"
                            " 13:15:38'")
        self.assertEqual(12, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("UPDATE all_datatypes SET a_char5= null,a_char20= "
                            "'a+34%sdf$sga$',a_varchar1= '',a_varchar50 = ?,"
                            "a_real = -3445.23423"
                            ",a_double= null,a_byteint= 55,a_smallint  = ?,"
                            "a_bigint= 45575616"
                            "557835435,a_numps= null,a_nump= ?,a_date= null,"
                            "a_time= '23:"
                            "45:56',a_timetz= ?,a_timestamp= '2005-12-31 "
                            "12:45:34',"
                            "a_interval= null", ('DGdfg56h67+5$-jgs24'
                                                 'j89vfd%Q', 23111,
                                                 12321, '12:22:23+10:30'))
        self.assertEqual(12, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_Update1(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("UPDATE all_datatypes SET a_date=?,a_time= ?,"
                            "a_timetz= ?, a_timestamp= ?, a_interval= ? "
                            "WHERE a_char5= ? AND a_varchar1= ? AND "
                            "a_real= ? AND a_smallint= ? AND a_bigint= "
                            "? AND a_numps= ? AND a_char20 LIKE ? AND "
                            "a_varchar50 LIKE ?", ('1901-10-15', '15:46:50',
                                                   '13:12:25+08:21',
                                                   '1997-06-18 10:15:25',
                                                   '22 years 13 months '
                                                   '35 days 13 hours '
                                                   '90 minutes 23 seconds',
                                                   'pe*o', 'x',
                                                   -6.41581e-10, -31380,
                                                   780406997461985024,
                                                   53.00000000000000,
                                                   'r8EQa6I/eP3l:2%',
                                                   'g.4+:_*/Y-066VKPLs'
                                                   'gQxnQ/f5'))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_BulkDelete(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("DELETE FROM all_datatypes WHERE a_char20 "
                            "LIKE ?", ('trial%',))
        self.assertEqual(2, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_Bulkdelete(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("DELETE FROM all_datatypes WHERE a_char5 "
                            "LIKE ? || "
                            "'%' AND a_varchar1 LIKE ?  AND a_srno = "
                            "8346 AND a_smallint = ?", ('S', 'c', -29092))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_InsertNull(self):
        self.cursor.execute("CREATE TABLE t1 (c1 float4, c2 double, c3 int1, "
                            "c4 int2, c5 int4, c6 int8, c7 char(5), "
                            "c8 varchar(10), c9 nchar(50), "
                            "c10 nvarchar(120),c11 varbinary(12), "
                            "c12 ST_GEOMETRY(12), DATE_PROD DATE, "
                            "TIME_PROD TIME, INTERVAL_PROD INTERVAL,"
                            "TIMESTMP TIMESTAMP,TIMETZ_PROD "
                            "TIME WITH TIME ZONE, "
                            "c18 bool)")
        self.cursor.execute("INSERT INTO t1 VALUES(?,?,?,?,?,?,?,?,"
                            "?,?,?,?,?,?,?,?,?,?)", (None, None, None,
                                                     None, None, None,
                                                     None, None, None,
                                                     None, None, None,
                                                     None, None, None,
                                                     None, None, None,))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("SELECT * FROM t1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, \
            c13, c14, c15, c16, c17, c18 \
                in results:
            self.assertEqual(None, c1, "ERROR: Data Difference")
            self.assertEqual(None, c2, "ERROR: Data Difference")
            self.assertEqual(None, c3, "ERROR: Data Difference")
            self.assertEqual(None, c4, "ERROR: Data Difference")
            self.assertEqual(None, c5, "ERROR: Data Difference")
            self.assertEqual(None, c6, "ERROR: Data Difference")
            self.assertEqual(None, c7, "ERROR: Data Difference")
            self.assertEqual(None, c8, "ERROR: Data Difference")
            self.assertEqual(None, c9, "ERROR: Data Difference")
            self.assertEqual(None, c10, "ERROR: Data Difference")
            self.assertEqual(None, c11, "ERROR: Data Difference")
            self.assertEqual(None, c12, "ERROR: Data Difference")
            self.assertEqual(None, c13, "ERROR: Data Difference")
            self.assertEqual(None, c14, "ERROR: Data Difference")
            self.assertEqual(None, c15, "ERROR: Data Difference")
            self.assertEqual(None, c16, "ERROR: Data Difference")
            self.assertEqual(None, c17, "ERROR: Data Difference")
            self.assertEqual(None, c18, "ERROR: Data Difference")

        self.cursor.execute("drop table t1 ")

    def test_InsertNullDate(self):
        self.cursor.execute("CREATE TABLE \"Mixed Case Crap Table\" "
                            "(\"a Mixed Case Column\" int, ts TIMESTAMP, "
                            "dt date, tm time, ch varchar(100))")
        self.cursor.execute("INSERT INTO \"Mixed Case "
                            "Crap Table\" (dt) VALUES(?)", (None,))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("SELECT * FROM \"Mixed Case Crap Table\" ")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5 in results:
            self.assertEqual(None, c1, "ERROR: Data Difference")
            self.assertEqual(None, c2, "ERROR: Data Difference")
            self.assertEqual(None, c3, "ERROR: Data Difference")
            self.assertEqual(None, c4, "ERROR: Data Difference")
            self.assertEqual(None, c5, "ERROR: Data Difference")

        self.cursor.execute("drop table \"Mixed Case Crap Table\" ")

    def test_InsertNumeric(self):
        self.cursor.execute("CREATE TABLE all_numeric (col1 numeric (36,14),"
                            "col2 numeric (38,0),col3 numeric (38,38),"
                            "col4 numeric (3,1),col5 numeric (18,10),"
                            "col6 numeric (38,1))")
        self.cursor.execute("INSERT INTO all_numeric values "
                            "(111111111111111111111.3333333333333,"
                            "2222222222222222222222222222222222222,"
                            "0.9999999999999999999999999999999999999,"
                            ".1,1111111.1111,"
                            "111111111111111111111111111111111111.1)")
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_numeric ")

    def test_MixedCaseInsert(self):
        self.cursor.execute("CREATE TABLE \"Mixed Case Crap Table\" (\"a "
                            "Mixed Case Column\" "
                            "int, ts TIMESTAMP, dt date, "
                            "tm time, ch varchar(100))")
        self.cursor.execute("INSERT INTO \"Mixed Case Crap Table\" "
                            "VALUES (?,?,?,?,?)", (1000000000, '1979-9-6',
                                                   '1979-9-6 10:10:10.123456',
                                                   '10:10:10', 'a\\nbc\\tdde'
                                                               '\\\\fgh\\bsi'
                                                               'jk\\rl'))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table \"Mixed Case Crap Table\" ")

    def test_Insert(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("INSERT INTO all_datatypes (a_srno, a_char5, "
                            "a_char20, a_char1000, a_real,a_double, "
                            "a_byteint, a_smallint, a_int, a_bigint,"
                            "a_numps, a_nump, a_dec, a_angle) VALUES "
                            "(2,'UR', 'q.+nZ', '',?, ?, ?,"
                            " ?, ?, ?, ?, ?, ?, ?)", (-4.19711e-35, 7.771,
                                                      -115, -24930,
                                                      869594835,
                                                      -16484452487417700,
                                                      3331.00000000000000,
                                                      9916, 3342.0000000,
                                                      9))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_AlterTable1(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("ALTER TABLE all_datatypes drop COLUMN "
                            "a_char5 cascade")
        self.cursor.execute("ALTER TABLE all_datatypes  ADD COLUMN "
                            "a_char5 char(5)")
        self.cursor.execute("INSERT INTO all_datatypes (a_srno, a_char5, "
                            "a_char20,a_char1000, "
                            "a_varchar1, a_varchar50) VALUES (1,'abcde',"
                            "'abcdefghijklmn', 'abcdefghijklmnop', "
                            "'a','abcd12356khmf')")
        self.cursor.execute("UPDATE all_datatypes  SET a_char5 = 'aaaaa'")
        self.assertEqual(13, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_DirectAggr(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT max (a_srno), max (a_char5), "
                            "max (a_char20), max (a_char1000), "
                            "max (a_varchar1), max (a_varchar50), "
                            "max (a_real), max (a_byteint), "
                            "max (a_smallint), max (a_bigint),"
                            "max (a_dec), max (a_angle), "
                            "max (a_date), max (a_time), "
                            "max(a_timetz), max (a_timestamp),"
                            "max(a_interval)"
                            " FROM all_datatypes ")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, \
            c11, c12, c13, c14, c15, c16, c17 \
                in results:
            self.assertEqual(8346, c1, "ERROR: Data Difference")
            self.assertEqual('xSgB ', c2, "ERROR: Data Difference")
            self.assertEqual('zNuB', c3, "ERROR: Data Difference")
            self.assertEqual('y', c5, "ERROR: Data Difference")
            self.assertEqual('25 years 67 days 18:10:05', c17,
                             "ERROR: Data Difference")
            self.assertEqual(907713502163946368, c10, "ERROR: Data Difference")
            self.assertEqual('2197-02-15', c13, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_RemoteUnloadBasic(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("CREATE EXTERNAL TABLE "
                            "all_datatypes_retunload_ext "
                            "(a_srno integer ,a_char5 char(5), "
                            "a_char20 char(20) "
                            "not NULL, a_char1000 char(1000), "
                            "a_varchar1 varchar(1), "
                            "a_varchar50 varchar(50), "
                            "a_real float(5), "
                            "a_double float(15), "
                            "a_byteint byteint, "
                            "a_smallint smallint, "
                            "a_int integer, "
                            "a_bigint bigint, "
                            "a_numps numeric(36,14), "
                            "a_nump numeric(5), "
                            "a_dec decimal(15,7), a_angle integer, "
                            "a_date date, "
                            "a_time time, a_timetz timetz, "
                            "a_timestamp timestamp, "
                            "a_interval interval ) USING "
                            "(DATAOBJECT "
                            "('/tmp/all_datatypes_retunload_ext_uncompressed')"
                            "REMOTESOURCE 'python' "
                            "DATESTYLE 'DMY' DELIMITER ',' "
                            "COMPRESS 'false' ENCODING 'internal')")
        self.cursor.execute("INSERT INTO all_datatypes_retunload_ext "
                            "SELECT * from all_datatypes")
        self.assertEqual(12, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes_retunload_ext")
        self.cursor.execute("drop table all_datatypes")

    def test_RemoteUnloadWithDelim(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("CREATE EXTERNAL TABLE "
                            "all_datatypes_retunload_ext "
                            "(a_srno integer ,a_char5  "
                            "char(5), a_char20 char(20) not NULL, "
                            "a_char1000 char(1000), a_varchar1 varchar(1), "
                            "a_varchar50 varchar(50), a_real float(5), "
                            "a_double float(15), a_byteint byteint, "
                            "a_smallint smallint, a_int integer, "
                            "a_bigint bigint, a_numps numeric(36,14), "
                            "a_nump numeric(5), a_dec decimal(15,7), "
                            "a_angle integer, a_date date, a_time time, "
                            "a_timetz timetz, a_timestamp "
                            "timestamp, a_interval interval ) "
                            "USING "
                            "(DATAOBJECT "
                            "('/tmp/all_datatypes_retunload_"
                            "ext_delim_default') "
                            "REMOTESOURCE "
                            "'python' DATESTYLE 'DMY' COMPRESS 'false' "
                            "ENCODING 'internal')")
        self.cursor.execute("INSERT INTO all_datatypes_retunload_ext "
                            "SELECT * from all_datatypes")
        self.assertEqual(12, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes_retunload_ext")
        self.cursor.execute("drop table all_datatypes")

    def test_RemoteUnloadFileSpec(self):
        self.performOperation("./setup_all_datatype")
        try:
            self.cursor.execute("CREATE EXTERNAL TABLE "
                                "all_datatypes_retunload_ext "
                                "(a_srno integer ,a_char5  char(5), "
                                "a_char20 char(20) not NULL, "
                                "a_char1000 char(1000), "
                                "a_varchar1 varchar(1), "
                                "a_varchar50 varchar(50), "
                                "a_real float(5), a_double float(15), "
                                "a_byteint byteint, a_smallint "
                                "smallint, a_int integer, a_bigint bigint, "
                                "a_numps numeric(36,14), "
                                "a_nump numeric(5), a_dec decimal(15,7), "
                                "a_angle integer, a_date date, "
                                "a_time time, a_timetz timetz, a_timestamp "
                                "timestamp, a_interval interval ) "
                                "USING (DATAOBJECT "
                                "('/tmp/all_datatypes_retunload_ext_rs_odbc') "
                                "REMOTESOURCE "
                                "'odbc' DELIMITER ',' DATESTYLE 'DMY' "
                                "COMPRESS 'false' ENCODING 'internal')")
            self.cursor.execute("INSERT INTO all_datatypes_retunload_ext "
                                "SELECT * from all_datatypes")
        except Exception as e:
            self.assertEqual(str(e), "ERROR:  Remotesource option "
                                     "of external table "
                                     "was not defined to load/unload "
                                     "using a python client\n\x00",
                             "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes_retunload_ext")
        self.cursor.execute("drop table all_datatypes")

    def test_RemoteUnloadBasicCompressed(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("CREATE EXTERNAL TABLE "
                            "all_datatypes_retunload_ext "
                            "(a_srno integer, a_char5 "
                            "char(5), a_char20 char(20) not NULL, "
                            "a_char1000 char(1000), "
                            "a_varchar1 varchar(1), a_varchar50 varchar(50), "
                            "a_real float(5), "
                            "a_double float(15), a_byteint byteint, "
                            "a_smallint smallint, "
                            "a_int integer, a_bigint bigint, "
                            "a_numps numeric(36,14), "
                            "a_nump numeric(5), a_dec decimal(15,7), "
                            "a_angle integer, "
                            "a_date date, a_time time, a_timetz timetz, "
                            "a_timestamp timestamp, "
                            "a_interval interval ) USING "
                            "(DATAOBJECT "
                            "('/tmp/all_datatypes_retunload_ext_compressed')"
                            "REMOTESOURCE 'python' COMPRESS 'true' FORMAT "
                            "'internal' ENCODING 'internal')")
        self.cursor.execute("INSERT INTO all_datatypes_retunload_ext "
                            "SELECT * from all_datatypes")
        self.assertEqual(12, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes_retunload_ext")
        self.cursor.execute("drop table all_datatypes")

    def test_Joins1(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT * FROM all_datatypes AS tab1 NATURAL JOIN "
                            "all_datatypes AS tab2 ORDER BY "
                            "tab1.a_srno, tab2.a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, \
            c12, c13, c14, c15, c16, c17, c18, c19, c20, c21 \
                in results:
            c = c + 1
        self.assertEqual(5, c, "ERROR: Data Difference")
        self.cursor.execute("SELECT * FROM all_datatypes AS tab1 INNER JOIN "
                            "all_datatypes AS tab2 using(a_srno) ORDER BY "
                            "tab1.a_srno, tab2.a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, \
            c11, c12, c13, c14, c15, c16, c17, c18, c19, c20, \
            c21, d1, d2, d3, d4, d5, d6, d7, d8, d9, d10, d11, \
            d12, d13, d14, d15, d16, d17, d18, d19, d20 \
                in results:
            c = c + 1
        self.assertEqual(12, c, "ERROR: Data Difference")
        self.cursor.execute("SELECT * FROM all_datatypes SELF JOIN "
                            "all_datatypes USING (a_srno)")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, \
            c13, c14, c15, c16, c17, c18, c19, c20, c21, d1, d2, \
            d3, d4, d5, d6, d7, d8, d9, d10, d11, d12, d13, d14, d15, \
            d16, d17, d18, d19, d20 \
                in results:
            c = c + 1
        self.assertEqual(12, c, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_Joins2(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT * FROM all_datatypes AS tab1 "
                            "RIGHT OUTER JOIN "
                            "all_datatypes AS tab2 USING(a_char20) ORDER BY "
                            "tab1.a_srno,tab2.a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, \
            c11, c12, c13, c14, c15, c16, c17, c18, c19, \
            c20, c21, d1, d2, d3, d4, d5, d6, d7, d8, d9, \
            d10, d11, d12, d13, d14, d15, d16, d17, d18, d19, d20 \
                in results:
            c = c + 1
        self.assertEqual(14, c, "ERROR: Data Difference")
        self.cursor.execute("SELECT * FROM all_datatypes AS tab1 LEFT OUTER "
                            "JOIN all_datatypes AS tab2 USING(a_char20) "
                            "ORDER BY tab1.a_srno,tab2.a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, \
            c12, c13, c14, c15, c16, c17, c18, c19, c20, c21, \
            d1, d2, d3, d4, d5, d6, d7, d8, d9, d10, d11, d12, d13, \
            d14, d15, d16, d17, d18, d19, d20 \
                in results:
            c = c + 1
        self.assertEqual(14, c, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_Joins3(self):
        self.performOperation("./setup_all_datatype")
        self.performOperation("./setup_all_latin")
        self.cursor.execute("SELECT * FROM all_Latin_datatypes, all_datatypes "
                            "ORDER BY all_latin_datatypes.a_srno,"
                            "all_datatypes.a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, c10,\
            c11, c12, c13, c14, c15, c16, c17, c18, c19, c20, \
            c21, d1, d2, d3, d4, d5 \
                in results:
            c = c + 1
        self.assertEqual(180, c, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")
        self.cursor.execute("drop table all_latin_datatypes")

    def test_DirectMixedComments(self):
        self.cursor.execute("CREATE TABLE test (int1 int, char1 char)")
        self.cursor.execute("INSERT INTO test VALUES /* "
                            "this is a comment */ (1, 'c')")
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("INSERT INTO test VALUES -- "
                            "this is a comment \n (2, 'a')")
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("INSERT INTO test VALUES /* this is "
                            "-- a comment */ (3, 'b')")
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table test ")

    def test_NumericFnEscSeq(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT abs(a_int), abs(a_smallint), "
                            "abs (a_bigint), "
                            "abs(a_double), abs(a_numps) FROM all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5 in results:
            self.assertEqual(798339977, c1, "ERROR: Data Difference")
            self.assertEqual(31380, c2, "ERROR: Data Difference")
            self.assertEqual(780406997461985024, c3, "ERROR: Data Difference")
            self.assertEqual(4242424422.60277, c4, "ERROR: Data Difference")
            self.assertEqual('53.00000000000000', c5, "ERROR: Data Difference")
        self.cursor.execute("SELECT round(a_numps), round(a_real), "
                            "round(a_double) "
                            "FROM all_datatypes ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3 in results:
            self.assertEqual('53', c1, "ERROR: Data Difference")
            self.assertEqual(0, c2, "ERROR: Data Difference")
            self.assertEqual(-4242424423, c3, "ERROR: Data Difference")

        self.cursor.execute("SELECT sqrt(a_srno), sqrt(a_nump) "
                            "FROM all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2 in results:
            self.assertEqual(7.615773105863909, c1, "ERROR: Data Difference")
            self.assertEqual(None, c2, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_SetOperator1(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT a_srno, a_char5, a_varchar1 "
                            "FROM all_datatypes "
                            "WHERE a_srno = ANY (SELECT a_srno "
                            "FROM all_datatypes "
                            "WHERE a_char1000 LIKE  ?) ORDER "
                            "BY a_srno", ('%',))
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3 in results:
            c = c + 1
        self.assertEqual(12, c, "ERROR: Data Difference")

        self.cursor.execute("SELECT a_srno, a_char5, a_date FROM "
                            "all_datatypes "
                            "WHERE a_srno = SOME (SELECT a_srno "
                            "FROM all_datatypes "
                            "WHERE a_date LIKE  ?) ORDER BY a_srno", ('%',))
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3 in results:
            c = c + 1
        self.assertEqual(11, c, "ERROR: Data Difference")

        self.cursor.execute("SELECT a_srno,a_varchar50 "
                            "FROM all_datatypes "
                            "a WHERE a_srno > ALL "
                            "(SELECT a_srno FROM all_datatypes "
                            "b WHERE a.a_srno <> b.a_srno)")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2 in results:
            c = c + 1
        self.assertEqual(1, c, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_SetOperator2(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT a_srno, a_char20 FROM "
                            "all_datatypes WHERE EXISTS "
                            "(SELECT a_srno FROM all_datatypes "
                            "WHERE a_char5 LIKE ?) "
                            "ORDER BY a_srno LIMIT 5", ('%',))
        results = self.cursor.fetchall()
        c = 0
        for c1, c2 in results:
            c = c + 1
        self.assertEqual(5, c, "ERROR: Data Difference")
        self.cursor.execute("SELECT a_srno,a_char20, a_interval "
                            "FROM all_datatypes "
                            "WHERE NOT EXISTS (SELECT a_char5 FROM "
                            "all_datatypes WHERE a_interval LIKE  ?) "
                            "ORDER BY a_srno DESC LIMIT 5",
                            ('25 years 67 days 18 hours 10 minutes',))
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3 in results:
            c = c + 1
        self.assertEqual(5, c, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_SetOperator3(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT a_srno, a_char20 FROM all_datatypes "
                            "WHERE a_char20 NOT IN (a_char5, "
                            "a_varchar50 )ORDER BY a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2 in results:
            c = c + 1
        self.assertEqual(11, c, "ERROR: Data Difference")
        self.cursor.execute("SELECT a_srno, a_time FROM all_datatypes a WHERE "
                            "a_srno IN (SELECT a_srno FROM all_datatypes b "
                            "WHERE a.a_srno = b.a_srno) ORDER BY a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2 in results:
            c = c + 1
        self.assertEqual(12, c, "ERROR: Data Difference")
        self.cursor.execute("SELECT a_srno, a_nump FROM "
                            "all_datatypes WHERE a_char5 "
                            "NOT IN (a_varchar1) AND a_srno "
                            "IN (SELECT a_srno FROM "
                            "all_datatypes WHERE a_nump  IS "
                            "NOT NULL) ORDER BY a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2 in results:
            c = c + 1
        self.assertEqual(9, c, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_SetOperator4(self):
        self.performOperation("./setup_all_latin")
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT a_srno, a_char20, "
                            "a_char1000 FROM all_datatypes "
                            "UNION SELECT a_srno, a_char10, "
                            "a_char1000 FROM all_latin_datatypes "
                            "WHERE a_srno IN (SELECT a_srno "
                            "FROM all_datatypes WHERE a_nump "
                            "IS NOT NULL)"
                            " ORDER BY a_srno")
        results = self.cursor.fetchall()
        c = 0
        for c1, c2, c3 in results:
            c = c + 1
        self.assertEqual(12, c, "ERROR: Data Difference")
        self.cursor.execute("SELECT * FROM ((SELECT a_char20 from "
                            "all_datatypes) "
                            "UNION ALL (SELECT a_char10 FROM "
                            "all_latin_datatypes)) AS a "
                            "ORDER BY 1")
        results = self.cursor.fetchall()
        expected = [[None], ['9'], ['BEEQxEnCuu-'], ['H5'],
                    ['HWFowZoZjwoXAj'], ['UZ3u2BocpErtN'],
                    ['gnXvyL'], ['q.+nZ'], ['r8EQa6I/eP3l:2RGR3tC'],
                    ['rbIx6*y'], ['trial'],
                    ['trial'], ['zNuB'], ['¢'], ['\xad'], ['·'], ['·ÃõýÉ'],
                    ['ÁãÑê'], ['ÄÖì'],
                    ['ÄÛºà'], ['Í»ü±'], ['ÕâÓÆ²ù¥'], ['ÖÔø'], ['Û\xadÃÐé°Ô'],
                    ['à'], ['àë'], ['òÛì']]
        output = []
        for c1 in results:
            output.append(c1)
        self.assertEqual(expected, output, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")
        self.cursor.execute("drop table ALL_LATIN_DATATYPES")

    def test_ExecuteScalar(self):
        self.performOperation("./setup_all_latin")
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT a_srno, a_char20, (SELECT a_char10 FROM "
                            "all_latin_datatypes WHERE "
                            "all_latin_datatypes.a_srno = 1) "
                            "FROM all_datatypes ORDER BY a_srno")
        results = self.cursor.fetchall()
        expected = [58, 'r8EQa6I/eP3l:2RGR3tC', 258, 'trial', 651,
                    'rbIx6*y', 698, 'BEEQxEnCuu-', 1628,
                    '9', 3336, 'q.+nZ', 5631, 'gnXvyL', 6324,
                    'UZ3u2BocpErtN', 6419, 'HWFowZoZjwoXAj',
                    6875, 'zNuB', 7253, 'H5', 8346, 'trial']
        output = []
        for c1, c2, c3 in results:
            output.append(c1)
            output.append(c2)
        self.assertEqual(expected, output, "ERROR: Data Difference")

        self.cursor.execute("SELECT a_srno, a_char20 FROM all_datatypes "
                            "WHERE 1 = (SELECT a_srno FROM "
                            "all_latin_datatypes "
                            "WHERE a_srno = 1)ORDER BY a_srno")
        results = self.cursor.fetchall()
        output = []
        for c1, c2 in results:
            output.append(c1)
            output.append(c2)
        self.assertEqual(expected, output, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")
        self.cursor.execute("drop table ALL_LATIN_DATATYPES")

    def test_FloatInIntCol(self):
        self.cursor.execute("CREATE TABLE test (col1 int)")
        self.cursor.execute("INSERT INTO test VALUES (.99)")
        self.cursor.execute("SELECT * from test")
        results = self.cursor.fetchall()
        for c in results:
            self.assertEqual([1], c, "ERROR: Data Difference")
        self.cursor.execute("drop table test")

    def test_FnEscSeq(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT ascii(a_srno), ascii(a_char20), "
                            "ascii(a_numps), "
                            "ascii(a_date), ascii(a_timestamp) "
                            "FROM all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5 in results:
            self.assertEqual(53, c1, "ERROR: Data Difference")
            self.assertEqual(114, c2, "ERROR: Data Difference")
            self.assertEqual(53, c3, "ERROR: Data Difference")
            self.assertEqual(49, c4, "ERROR: Data Difference")
            self.assertEqual(49, c5, "ERROR: Data Difference")

        self.cursor.execute("SELECT position(1 IN a_srno), "
                            "position(1 IN a_time), "
                            "position(1 IN a_interval) FROM all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3 in results:
            self.assertEqual(0, c1, "ERROR: Data Difference")
            self.assertEqual(1, c2, "ERROR: Data Difference")
            self.assertEqual(24, c3, "ERROR: Data Difference")

        self.cursor.execute("SELECT repeat(a_srno, 2), repeat(a_date, -1), "
                            "repeat(a_numps, 3), "
                            "repeat(a_char5, 2) FROM all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual('5858', c1, "ERROR: Data Difference")
            self.assertEqual('', c2, "ERROR: Data Difference")
            self.assertEqual('53.0000000000000053.0000000000000053.'
                             '00000000000000', c3, "ERROR: Data Difference")
            self.assertEqual('pe*o pe*o ', c4, "ERROR: Data Difference")

        self.cursor.execute("SELECT substring(a_char20, 2, 10), "
                            "substring(a_numps, 2, 10), "
                            "substring(a_date, 2, 15), substring"
                            "(a_varchar1, 1, 15) "
                            "FROM all_datatypes ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual('8EQa6I/eP3', c1, "ERROR: Data Difference")
            self.assertEqual('3.00000000', c2, "ERROR: Data Difference")
            self.assertEqual('982-04-22', c3, "ERROR: Data Difference")
            self.assertEqual('x', c4, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_ForCast1(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT a_srno,CAST (a_date as timestamp),"
                            "CAST (a_date as char(20)),"
                            "CAST (a_date as varchar(20))FROM all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual(58, c1, "ERROR: Data Difference")
            self.assertEqual('1982-04-22 00:00:00.000000', c2,
                             "ERROR: Data Difference")
            self.assertEqual('1982-04-22', c3, "ERROR: Data Difference")
            self.assertEqual('1982-04-22', c4, "ERROR: Data Difference")

        self.cursor.execute("SELECT a_srno,CAST (a_time as interval),CAST "
                            "(a_time as char(20)),"
                            "CAST (a_time as varchar(20))FROM all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual(58, c1, "ERROR: Data Difference")
            self.assertEqual('13:20:05', c2, "ERROR: Data Difference")
            self.assertEqual('13:20:05', c3, "ERROR: Data Difference")
            self.assertEqual('13:20:05', c4, "ERROR: Data Difference")

        self.cursor.execute("SELECT a_srno,CAST (a_interval as time), "
                            "CAST (a_interval as char(50)),"
                            "CAST (a_interval as varchar(50))FROM "
                            "all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual(58, c1, "ERROR: Data Difference")
            self.assertEqual('00:05:10', c2, "ERROR: Data Difference")
            self.assertEqual('20 years 67 days 00:05:10', c3,
                             "ERROR: Data Difference")
            self.assertEqual('20 years 67 days 00:05:10', c4,
                             "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_ForCast2(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT a_srno,CAST (a_char20 as char(2)),"
                            "CAST (a_char20 as varchar(20)),"
                            "CAST (a_varchar50 as char(50)),CAST "
                            "(a_varchar50 as char(5)) "
                            "FROM all_datatypes ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5 in results:
            self.assertEqual(58, c1, "ERROR: Data Difference")
            self.assertEqual('r8', c2, "ERROR: Data Difference")
            self.assertEqual('r8EQa6I/eP3l:2RGR3tC', c3,
                             "ERROR: Data Difference")
            self.assertEqual('g.4+:_*/Y-066VKPLsgQxnQ/f5', c4,
                             "ERROR: Data Difference")
            self.assertEqual('g.4+:', c5, "ERROR: Data Difference")

        self.cursor.execute("SELECT a_srno,CAST (a_real as double),"
                            "CAST (a_real as char (30)),"
                            "CAST (a_real as varchar (30))FROM all_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual(58, c1, "ERROR: Data Difference")
            self.assertEqual(-6.415810105409037e-10, c2, "ERROR: "
                                                         "Data Difference")
            self.assertEqual('-6.41581e-10', c3, "ERROR: Data Difference")
            self.assertEqual('-6.41581e-10', c4, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_ForCast3(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT CAST (a_angle as byteint),CAST "
                            "(a_srno as smallint),"
                            "CAST (a_srno as bigint),CAST (a_srno as real),"
                            "CAST (a_angle as numeric (10,2)),CAST "
                            "(a_angle as char(5)),"
                            "CAST (a_angle as varchar(5)) FROM all_datatypes"
                            " ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5, c6, c7 in results:
            self.assertEqual(-1, c1, "ERROR: Data Difference")
            self.assertEqual(58, c2, "ERROR: Data Difference")
            self.assertEqual(58, c3, "ERROR: Data Difference")
            self.assertEqual(58, c4, "ERROR: Data Difference")
            self.assertEqual('-1.00', c5, "ERROR: Data Difference")
            self.assertEqual('-1   ', c6, "ERROR: Data Difference")
            self.assertEqual('-1', c7, "ERROR: Data Difference")

        self.cursor.execute("SELECT CAST (a_numps as int), "
                            "CAST (a_numps as smallint), "
                            "CAST (a_numps as bigint), CAST "
                            "(a_numps as real), "
                            "CAST (a_numps as numeric (4,0)), "
                            "CAST (a_numps as char(30)),"
                            " CAST (a_numps as varchar(30)) FROM "
                            "all_datatypes ORDER BY a_srno "
                            "limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5, c6, c7 in results:
            self.assertEqual(53, c1, "ERROR: Data Difference")
            self.assertEqual(53, c2, "ERROR: Data Difference")
            self.assertEqual(53, c3, "ERROR: Data Difference")
            self.assertEqual(53, c4, "ERROR: Data Difference")
            self.assertEqual('53', c5, "ERROR: Data Difference")
            self.assertEqual('53.00000000000000', c6, "ERROR: Data Difference")
            self.assertEqual('53.00000000000000', c7, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_ForLikeClause(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT a_srno, a_char5, a_date, a_timetz FROM "
                            "all_datatypes WHERE a_char5 LIKE '%%%%' AND"
                            " a_date LIKE  '1977______' AND a_timetz like"
                            " '%+%' ORDER BY a_srno limit 1 ")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual(6419, c1, "ERROR: Data Difference")
            self.assertEqual('xSgB ', c2, "ERROR: Data Difference")
            self.assertEqual('1977-05-10', c3, "ERROR: Data Difference")
            self.assertEqual('23:40:05+05:30', c4, "ERROR: Data Difference")

        self.cursor.execute("UPDATE all_datatypes SET a_char20 = "
                            "('%' || a_char20) WHERE a_srno = 8346")
        self.cursor.execute("SELECT a_srno,a_char20, a_char1000 FROM "
                            "all_datatypes "
                            "WHERE a_char20 LIKE '\\%%' escape '\\' "
                            "AND a_char1000 LIKE '%_%'")
        results = self.cursor.fetchall()
        for c1, c2, c3 in results:
            self.assertEqual(8346, c1, "ERROR: Data Difference")
            self.assertEqual('%trial', c2, "ERROR: Data Difference")
            self.assertEqual('0_ZX2lJLkyeM4Mn6PgEEh8QFfbQMw8obO jKAq55BaKHJgWb'
                             'c*S9YJ74ih2pdd/AbjPS/angwdovN5Q*2N-IZY '
                             'pQSiLTj:e8dHFq9Qk'
                             '8Nskc2*F31WV-.AcO9_A5dFmfteO79Qb:u+v:kEr'
                             'AdyonLMNzr/x6fN'
                             '1rKV6qPxoR. pUOQZD*:7zQSYh/WvAN ullT5RaWu'
                             '/8 a9g_XetY4+tz/'
                             'TYD75g.C+_Q+1r+SfVEsmh53Tew2Tj_qEry_srWs'
                             'k1PCNdqlEvGDrBFqw'
                             'zorZxRw/Kxr9VHuBQZJX4ss0XNsk.mMu_4KdOl'
                             '9CXfAUjGxnr6UC'
                             'kGxN uUY-492S+xTQ9Qn0ZSfuNBVPUJtPII'
                             'c4:Ca0VSR2:JbL4.2V*_qGfRI4I',
                             c3, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_SysFn(self):
        self.cursor.execute("SELECT current_catalog,current_user,"
                            "coalesce(NULL,10)")
        results = self.cursor.fetchall()
        for c1, c2, c3 in results:
            self.assertEqual('NZPY_TEST', c1, "ERROR: Data Difference")
            self.assertEqual('ADMIN', c2, "ERROR: Data Difference")
            self.assertEqual(10, c3, "ERROR: Data Difference")

    def test_StringFn1(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("select ascii(a_varchar1) ,"
                            "(octet_length(a_varchar1) * 8),"
                            "chr(a_srno) ,char_length(a_varchar50) ,"
                            "char_length(a_varchar50) ,"
                            "varcharcat(a_varchar1,a_varchar50) ,"
                            "lower(a_varchar1) ,"
                            "substring(a_varchar50,1,4) ,"
                            "length(a_varchar50) ,"
                            "position('m' in a_varchar50),"
                            "ltrim(a_varchar50) ,"
                            "octet_length(a_varchar50) ,"
                            "position('g' in a_varchar50) ,"
                            "repeat(a_varchar1,2),"
                            "ltrim(a_varchar50) ,lpad('',a_srno) ,"
                            "substring(a_varchar50,2,5) ,upper(a_varchar50) "
                            "from all_datatypes order by a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5, c6, c7, c8, c9, \
            c10, c11, c12, c13, c14, c15, c16, c17, c18 \
                in results:
            self.assertEqual(120, c1, "ERROR: Data Difference")
            self.assertEqual(8, c2, "ERROR: Data Difference")
            self.assertEqual(':', c3, "ERROR: Data Difference")
            self.assertEqual(26, c4, "ERROR: Data Difference")
            self.assertEqual(26, c5, "ERROR: Data Difference")
            self.assertEqual('xg.4+:_*/Y-066VKPLsgQxnQ/f5',
                             c6, "ERROR: Data Difference")
            self.assertEqual('x', c7, "ERROR: Data Difference")
            self.assertEqual('g.4+', c8, "ERROR: Data Difference")
            self.assertEqual(26, c9, "ERROR: Data Difference")
            self.assertEqual(0, c10, "ERROR: Data Difference")
            self.assertEqual('g.4+:_*/Y-066VKPLsgQxnQ/f5',
                             c11, "ERROR: Data Difference")
            self.assertEqual(26, c12, "ERROR: Data Difference")
            self.assertEqual(1, c13, "ERROR: Data Difference")
            self.assertEqual('xx', c14, "ERROR: Data Difference")
            self.assertEqual('g.4+:_*/Y-066VKPLsgQxnQ/f5',
                             c15, "ERROR: Data Difference")
            self.assertEqual('                                   '
                             '                       ', c16,
                             "ERROR: Data Difference")
            self.assertEqual('.4+:_', c17, "ERROR: Data Difference")
            self.assertEqual('G.4+:_*/Y-066VKPLSGQXNQ/F5',
                             c18, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_CountColumnWithMarker(self):
        self.performOperation("./setup_all_datatype")
        self.cursor.execute("SELECT COUNT(*) FROM all_datatypes "
                            "WHERE a_srno < ?", (1000,))
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual([4], c1, "ERROR: Data Difference")
        self.cursor.execute("SELECT COUNT(*) FROM all_datatypes"
                            " WHERE a_srno > ? ORDER BY COUNT(*) "
                            "LIMIT 1", (6419,))
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual([3], c1, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_TemporalDatatypes1(self):
        self.cursor.execute("SELECT date_part('day',2007-07-25)")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual([1], c1, "ERROR: Data Difference")

        self.cursor.execute("SELECT date_part('dow',2007-07-25)")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual([5], c1, "ERROR: Data Difference")

        self.cursor.execute("SELECT date_part('doy',2007-07-25),"
                            "date_part('MONTH',2007-07-25)")
        results = self.cursor.fetchall()
        for c1, c2 in results:
            self.assertEqual(1, c1, "ERROR: Data Difference")
            self.assertEqual(1, c2, "ERROR: Data Difference")

        self.cursor.execute("SELECT date_part('QUARTER',2007-07-25), "
                            "date_part('WEEK',2007-07-25),date_part"
                            "('YEAR',2007-07-25)")
        results = self.cursor.fetchall()
        for c1, c2, c3 in results:
            self.assertEqual(1, c1, "ERROR: Data Difference")
            self.assertEqual(1, c2, "ERROR: Data Difference")
            self.assertEqual(1970, c3, "ERROR: Data Difference")

    def test_TemporalDatatypes2(self):
        self.cursor.execute("CREATE TABLE tab (x integer,a date) ")
        self.cursor.execute("insert into tab  values (1, '2008-09-09')")
        self.cursor.execute("insert into tab  values (2, '1999-09-19')")
        self.cursor.execute("insert into tab  values (3, '2005-07-09')")
        self.cursor.execute("insert into tab  values (4, '1999-09-09')")
        self.cursor.execute("select x from tab where (date_part('YEAR',a) "
                            "= 1999) and (date_part('MONTH',a) = 9) and "
                            "(date_part('day',a) = 9)")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual([4], c1, "ERROR: Data Difference")
        self.cursor.execute("drop table tab")

    def test_TemporalDatatypes3(self):
        self.cursor.execute("CREATE TABLE test1 (dateCol date) ")
        self.cursor.execute("Insert into test1 VALUES ('1977-10-05 BC')")
        self.cursor.execute("SELECT * FROM test1")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual(['-1976-10-05'], c1, "ERROR: Data Difference")
        self.cursor.execute("DELETE FROM test1")
        self.cursor.execute("INSERT INTO test1 VALUES ('9999-12-30')")
        self.cursor.execute("SELECT  (dateCol + INTERVAL "
                            "'@ 1DAY') AS next_date FROM test1")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual(['9999-12-31 00:00:00.0000'
                              '00'], c1, "ERROR: Data Difference")
        self.cursor.execute("DELETE FROM test1")
        self.cursor.execute("INSERT INTO test1 VALUES ('48010301 BC')")
        self.cursor.execute("SELECT * FROM test1")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual(['-4800-03-01'], c1, "ERROR: Data Difference")
        self.cursor.execute("drop table test1")

    def test_LimitwithOffset(self):
        self.performOperation("./setup_all_datatype")
        expected = [[651], [698], [1628], [3336], [5631]]
        output = []
        self.cursor.execute("SELECT a_srno from all_datatypes "
                            "order by a_srno LIMIT 5 ,2")
        results = self.cursor.fetchall()
        for c1 in results:
            output.append(c1)
        self.assertEqual(output, expected, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_BoolCol(self):
        self.cursor.execute("CREATE TABLE all_datatypes (a_srno "
                            "integer,a_bool5 bool )")
        self.cursor.execute("Insert into all_datatypes values(10,'t')")
        self.cursor.execute("Insert into all_datatypes values(5,'0')")
        self.cursor.execute("Insert into all_datatypes values(12,'1')")
        self.cursor.execute("Insert into all_datatypes values(3,'f')")
        self.cursor.execute("SELECT * FROM all_datatypes "
                            "ORDER BY a_srno limit 1;")
        results = self.cursor.fetchall()
        for c1, c2 in results:
            self.assertEqual(3, c1, "ERROR: Data Difference")
            self.assertEqual(0, c2, "ERROR: Data Difference")
        self.cursor.execute("SELECT * FROM all_datatypes "
                            "ORDER BY a_srno DESC limit 1;")
        results = self.cursor.fetchall()
        for c1, c2 in results:
            self.assertEqual(12, c1, "ERROR: Data Difference")
            self.assertEqual(1, c2, "ERROR: Data Difference")
        self.cursor.execute("drop table all_datatypes")

    def test_Varbinary(self):
        self.cursor.execute("create table varbin_test (a_srno "
                            "int,col0 VARBINARY(50));")
        self.cursor.execute("insert into varbin_test values(?, "
                            "?)", (1, b'543b6c6c6f'))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("select col0 from varbin_test")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual(['T;llo'], c1, "ERROR: Data Difference")
        self.cursor.execute("update varbin_test set col0 = ? where a_srno = "
                            "?", (b'543b6c6c6a', 1))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("select col0 from varbin_test")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual(['T;llj'], c1, "ERROR: Data Difference")
        self.cursor.execute("delete from varbin_test where "
                            "col0 = ?", (b'543b6c6c6a',))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table varbin_test")

    def test_Geometry(self):
        self.cursor.execute("create table geometry_test (a_srno int,"
                            "col0 ST_GEOMETRY(50));")
        self.cursor.execute("insert into geometry_test values(?, "
                            "?)", (1, b'123d6c6f'))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("select col0 from geometry_test")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual(['\x12=lo'], c1, "ERROR: Data Difference")
        self.cursor.execute("update geometry_test set col0 = ? where "
                            "a_srno = ?", (b'123d6c6a', 1))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("select col0 from geometry_test")
        results = self.cursor.fetchall()
        for c1 in results:
            self.assertEqual(['\x12=lj'], c1, "ERROR: Data Difference")
        self.cursor.execute("delete from geometry_test where col0 = ?",
                            (b'123d6c6a',))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table geometry_test")

    def test_ForFetchAllDataUtf8(self):
        self.performOperation("./setup_all_utf")
        self.cursor.execute("SELECT * FROM all_utf_datatypes ORDER BY "
                            "a_srno DESC limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual(15, c1, "ERROR: Data Difference")
            self.assertEqual('􃩝虲􄃨𨘊󻻶􎮟', c2, "ERROR: Data Difference")
            self.assertEqual('􆴩􁁭ꏫ󾆲𤍈󰘋󰘑', c4, "ERROR: Data Difference")
        self.cursor.execute("drop table all_utf_datatypes")

    def test_DeletetaUtf8(self):
        self.performOperation("./setup_all_utf")
        self.cursor.execute("delete from all_utf_datatypes where "
                            "a_nchar10 = ?", ('䄕莆',))
        self.assertEqual(1, self.cursor.rowcount, "ERROR: Data Difference")
        self.cursor.execute("drop table all_utf_datatypes")

    def test_Literals(self):
        self.performOperation("./setup_all_latin")
        self.cursor.execute("SELECT 'ODBC' || 'Testing' as test, "
                            "'Latin9' || a_char10, "
                            "'Netezza' a_char1000, 'Persistent' FROM "
                            "all_latin_datatypes "
                            "WHERE a_char1000 like  ? ORDER BY a_char10 "
                            "DESC limit 1", ('%Ã%',))
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual('ODBCTesting', c1, "ERROR: Data Difference")
            self.assertEqual('Latin9òÛì       ', c2, "ERROR: Data Difference")
            self.assertEqual('Netezza', c3, "ERROR: Data Difference")
            self.assertEqual('Persistent', c4, "ERROR: Data Difference")

        self.cursor.execute("SELECT 'Test Literals ',"
                            "(((((a_srno * 10 ) + 20 ) - 10 ) / 10)),"
                            "'complex operation' FROM all_latin_datatypes "
                            "ORDER BY a_char10 limit 1", ('%Ã%',))
        results = self.cursor.fetchall()
        for c1, c2, c3 in results:
            self.assertEqual('Test Literals ', c1, "ERROR: Data Difference")
            self.assertEqual(7, c2, "ERROR: Data Difference")
            self.assertEqual('complex operation', c3, "ERROR: Data Difference")

        self.cursor.execute("drop table all_latin_datatypes")

    def test_LatinAggregates(self):
        self.performOperation("./setup_all_latin")
        self.cursor.execute("SELECT count(*),count(a_char10) FROM "
                            "all_latin_datatypes;")
        results = self.cursor.fetchall()
        for c1, c2 in results:
            self.assertEqual(15, c1, "ERROR: Data Difference")
            self.assertEqual(14, c2, "ERROR: Data Difference")

        self.cursor.execute("SELECT count(*),count(a_char1000), "
                            "count(a_varchar10), count(a_varchar32767) "
                            "FROM all_latin_datatypes;")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual(15, c1, "ERROR: Data Difference")
            self.assertEqual(15, c2, "ERROR: Data Difference")
            self.assertEqual(15, c3, "ERROR: Data Difference")
            self.assertEqual(15, c4, "ERROR: Data Difference")

        self.cursor.execute("SELECT max (a_srno),max (a_char10), "
                            "min (a_srno), min (a_char10) "
                            "FROM all_latin_datatypes;")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4 in results:
            self.assertEqual(15, c1, "ERROR: Data Difference")
            self.assertEqual('òÛì       ', c2, "ERROR: Data Difference")
            self.assertEqual(1, c3, "ERROR: Data Difference")
            self.assertEqual('¢         ', c4, "ERROR: Data Difference")

        self.cursor.execute("drop table all_latin_datatypes")

    def test_UtfAggregates(self):
        self.performOperation("./setup_all_utf")
        self.cursor.execute("SELECT max (a_srno), max (a_nchar10), "
                            "max (a_nvarchar10), min (a_srno), "
                            "min (a_nchar10), "
                            "min (a_nvarchar10) FROM all_utf_datatypes;")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5, c6 in results:
            self.assertEqual(15, c1, "ERROR: Data Difference")
            self.assertEqual('􏜔󾄌', c2, "ERROR: Data Difference")
            self.assertEqual('􆴩􁁭ꏫ󾆲𤍈󰘋󰘑', c3, "ERROR: Data Difference")
            self.assertEqual(1, c4, "ERROR: Data Difference")
            self.assertEqual('', c5, "ERROR: Data Difference")
            self.assertEqual('埇䕼姲􁕢􅳶', c6, "ERROR: Data Difference")

        self.cursor.execute("SELECT max (a_srno),max (a_nchar1000) "
                            "FROM all_utf_datatypes;")
        results = self.cursor.fetchall()
        for c1, c2 in results:
            self.assertEqual(15, c1, "ERROR: Data Difference")
            self.assertEqual('􌘜乂畾𦼡􎵱􃥥􄬘􄠋󲼱𨗽𩌍􏍰󴓃󰆋𩨗뿡𦏠𩍨𧏏뗒􇷟󱱓􈩺􉁱󺻅픫璸󳃄𣢦󷙫󲢃󶼐󿌸􄎺蔤噮􀳉ꎸㅓ􇍦', c2,
                             "ERROR: Data Difference")

        self.cursor.execute("SELECT min (a_srno),min (a_nchar1000) "
                            "FROM all_utf_datatypes;")
        results = self.cursor.fetchall()
        for c1, c2 in results:
            self.assertEqual(1, c1, "ERROR: Data Difference")
            self.assertEqual('', c2, "ERROR: Data Difference")

        self.cursor.execute("drop table all_utf_datatypes")

    def test_ConcatLatin(self):
        self.performOperation("./setup_all_latin")
        self.cursor.execute("SELECT varcharcat(a_char10, a_varchar10), "
                            "varcharcat(a_varchar10, a_char10), "
                            "varcharcat(a_char10, a_char10) FROM "
                            "all_latin_datatypes "
                            "ORDER BY a_srno limit 1")
        results = self.cursor.fetchall()
        for c1, c2, c3 in results:
            self.assertEqual('¢         Ó£ì', c1, "ERROR: Data Difference")
            self.assertEqual('Ó£ì¢         ', c2, "ERROR: Data Difference")
            self.assertEqual('¢         ¢       '
                             '  ', c3, "ERROR: Data Difference")

        self.cursor.execute("drop table all_latin_datatypes")

    def test_FetchLatinUtf(self):
        self.performOperation("./setup_all_latin_utf")
        self.cursor.execute("select * from all_utfLatin_datatypes")
        results = self.cursor.fetchall()
        for c1, c2, c3, c4, c5, c6, c7 in results:
            self.assertEqual(1, c1, "ERROR: Data Difference")
            self.assertEqual('àë        ', c2, "ERROR: Data Difference")
            self.assertEqual('©¡ùâûôÞçå¶ý§ë¡êÆù×Òþ®²¬óßöÅè§ýÙ'
                             'åèáÓ¥ë³àûñ¥²ïú', c3, "ERROR: Data Difference")
            self.assertEqual('\xa0ì«¹Ü', c4,
                             "ERROR: Data Difference")
            self.assertEqual('î·®ð£', c5, "ERROR: Data Difference")
            self.assertEqual('ó´<9a><90>ð£¨°ð¥»<85>ô<8a>°¦ð¨<88>¾ó·¿<90>óµ'
                             '¿¡ô<85><8c>«ð<9d'
                             '><9b>»ô<8c>³<81>ê<8b>ªð <89>¨ô<8b><82><93>ô'
                             '<8c>½<9a>æ¢<8b>è°<9c>ô<', c6,
                             "ERROR: Data Difference")
            self.assertEqual('å', c7, "ERROR: Data Difference")
        self.cursor.execute("drop table all_utfLatin_datatypes")



import ctypes

NDIGIT_INT64 = False
MAX_NUMERIC_DIGIT_COUNT = 4
NUMERIC_MAX_PRECISION = 38
HI32_MASK = 0xffffffff00000000
USE_MUL_DOUBLE = False
const_data_ten = []


class NumericVar():

    def __init__(self):
         self.data           = []
         self.scale          = None
         self.rscale         = None
         self.rprecision     = None

def int32_to_uint32(i):
    return ctypes.c_uint32(i).value

def base():
    return (1 << 32)

def highPart(val):
    return (val >> 32)

def lowPart(val):
    return (val & (1 << 32) - 1)

def encodeNum(words, low, hi):
    words[0] = int(lowPart(low))
    words[1] = int(highPart(low))
    words[2] = int(lowPart(hi))
    words[3] = int(highPart(hi))

#TODO: this function ultimately does nothing. Either remove it or redesign it.
def decodeNum(words, low, hi):
    low = words[0] | words[1]*(1<<32)
    hi = words[2] | words[3]*(1<<32)

def isNumeric_Data_Negative(numdataP):
    return (numdataP[0] & 0x80000000) != 0

def copy_128(srcP):
    destP = []
    destP.append(srcP[0])
    destP.append(srcP[1])
    destP.append(srcP[2])
    destP.append(srcP[3])
    return destP

def negate_128(arg):
    #First complement the value (1's complement)
    for i in range(MAX_NUMERIC_DIGIT_COUNT):
        arg[i] = int32_to_uint32(~arg[i])

    #Then increment it to form 2's complement (negative)
    return inc_128(arg)

#for 2's complement
def inc_128(arg ):
    i = MAX_NUMERIC_DIGIT_COUNT
    carry = True
    bInputNegative = isNumeric_Data_Negative(arg)

    while (i != 0) and carry:
        i -= 1
        work = (arg[i]) + 1
        carry = (work & HI32_MASK) != 0
        arg[i] = work & 0xffffffff

    if not bInputNegative:
        return isNumeric_Data_Negative(arg)
    else:
        return False

def div_128(numeratorP, denominatorP, resultP):
    hiquotient = 0
    loquotient = 0
    hiremainder = 0
    loremainder = 0
    bResNegative = isNumeric_Data_Negative(numeratorP) != isNumeric_Data_Negative(denominatorP)

    num = copy_128(numeratorP)
    if isNumeric_Data_Negative(num):
        if negate_128(num):
            return True

    den = copy_128(denominatorP)
    if isNumeric_Data_Negative(den):
        if negate_128(den):
            return True

    hinum = ((num[0]) << 32) + (num[1])
    lonum = ((num[2]) << 32) + (num[3])
    hidenom = ((den[0]) << 32) + (den[1])
    lodenom = ((den[2]) << 32) + (den[3])

    if div_and_round_double(1, lonum, hinum, lodenom, hidenom, loquotient, hiquotient, loremainder, hiremainder) != 0:
        return True

    resultP[0] = (((hiquotient) & HI32_MASK) >> 32)
    resultP[1] = (hiquotient & 0xffffffff)
    resultP[2] = (((loquotient) & HI32_MASK) >> 32)
    resultP[3] = (loquotient & 0xffffffff)

    if bResNegative:
        if negate_128(resultP):
            return True

    return False

def div10_128(numeratorP, quotientP):
    remainder = 0

    for i in range(MAX_NUMERIC_DIGIT_COUNT):
        work = numeratorP[i] + (remainder<<32)
        if work != 0:
            quotientP[i] = int(work / 10)
            remainder = int(work % 10)
        else:
            quotientP[i] = 0
            remainder = 0
    return remainder

def PYTHON_numeric_load_var(dataP, precision, scale, digitCount):
    varP = NumericVar()

    #extend sign
    sign = dataP[0] & 0x80000000
    if sign != 0:
        leadDigit = 0xffffffff
    else:
        leadDigit = 0

    i = 0
    while i < MAX_NUMERIC_DIGIT_COUNT-digitCount:
        varP.data.append(leadDigit)
        i +=1

    j = 0
    while i < MAX_NUMERIC_DIGIT_COUNT:
        varP.data.append(dataP[j])
        j +=1
        i +=1

    varP.scale = scale
    varP.rscale = scale
    varP.rprecision = precision

    return varP

# ----------
#  round_var() -
#       Rounds a numeric var to a target scale.
#  overflow will return true.
# ----------
def round_var(nvar, scale):
    for _ in range(MAX_NUMERIC_DIGIT_COUNT):
        const_data_ten.append(0)

    const_data_ten[MAX_NUMERIC_DIGIT_COUNT-1] = 10

    aDD = scale - nvar.scale # additional decimal digits
    positive = not isNumeric_Data_Negative(nvar.data)
    temp = []

    if nvar.scale == scale:
        return False

    workData = copy_128(nvar.data)
    if not positive:
        if negate_128(workData):
            return True

    if aDD < 0:
        if (aDD != -1) and div_128(workData, power_of_10(-aDD-1), workData):
            return True

        round = (div10_128(workData, temp) > 4) #for rounding ro next number
        if div_128(workData, const_data_ten, workData):
            return True

        if round:
            if inc_128(workData):
                return True

    elif aDD > 0:
        if mul_128(workData, power_of_10(aDD), workData):
            return True

    nvar.scale = scale
    nvar.rscale = scale # !FIX-jpb   rounding should change result scale, right?
    nvar.data = copy_128(workData)
    if not positive:
        if negate_128(nvar.data):
            return True

    return False

# Multiply two doubleword integers with doubleword result.
# Return nonzero if the operation overflows, assuming it's signed.
# Each argument is given as two `int64' pieces.
# One argument is L1 and H1; the other, L2 and H2.
# The value is stored as two `int64' pieces in *LV and *HV.
def mul_double(l1, h1, l2, h2, lv, hv):
    arg1 = []
    arg2 = []
    prod = [0, 0, 0, 0, 0, 0, 0, 0]
    toplow = 0
    tophigh = 0
    neglow = 0
    neghigh = 0

    encodeNum(arg1, l1, h1)
    encodeNum(arg2, l2, h2)

    for i in range(4):
        carry = 0
        for j in range(4):
            k = i + j
            # This product is <= 0xFFFE0001, the sum <= 0xFFFF0000.
            carry += arg1[i] * arg2[j]
            # Since prod[p] < 0xFFFF, this sum <= 0xFFFFFFFF.
            carry += prod[k]
            prod[k] = lowPart(carry)
            carry = highPart(carry)

        prod[i+4] = carry

    decodeNum(prod[:4], lv, hv) # This ignores prod[4] through prod[4*2-1]

    # Check for overflow by calculating the top half of the answer in full;
    # it should agree with the low half's sign bit.
    decodeNum(prod[4:], toplow, tophigh)
    if h1 < 0:
        neg_double(l2, h2, neglow, neghigh)
        add_double(neglow, neghigh, toplow, tophigh, toplow, tophigh)

    if h2 < 0:
        neg_double(l1, h1, neglow, neghigh)
        add_double(neglow, neghigh, toplow, tophigh, toplow, tophigh)

    if hv < 0:
        return ((~(toplow & tophigh)) != 0)
    else:
        return ((toplow | tophigh) != 0)

# Negate a doubleword integer with doubleword result.
# Return nonzero if the operation overflows, assuming it's signed.
# The argument is given as two `int64' pieces in L1 and H1.
# The value is stored as two `int64' pieces in *LV and *HV.
def neg_double(l1 , h1, lv, hv):
    if l1 == 0:
        lv = 0 #TODO: value isn't used anywhere. Revise or remove.
        hv = -h1
        if (hv & h1) < 0:
            return 1
        else:
            return 0
    else:
        lv = -l1 #TODO: value isn't used anywhere. Revise or remove.
        hv = ~h1
        return 0

def overflow_sum_sign(a, b, sum):
    if (~(a ^ b) & (a ^ sum)) < 0:
        return 1
    else:
        return 0

# Add two doubleword integers with doubleword result.
# Each argument is given as two `int64' pieces.
# One argument is L1 and H1; the other, L2 and H2.
# The value is stored as two `int64' pieces in *LV and *HV.
def add_double(l1, h1, l2, h2, lv, hv):
    l = l1 + l2

    if l < l1:
        badd = 1
    else:
        badd = 0

    h = h1 + h2 + badd

    lv = l #TODO: value isn't used anywhere. Revise or remove.
    hv = h #TODO: value isn't used anywhere. Revise or remove.

    return overflow_sum_sign(h1, h2, h)

#Get Numeric variable value represented in string format
def get_str_from_var(nvar, dscale):
    res = ""
    unbiasedDigits = []
    work = []
    bLeadingZeroes = True
    pos = 0

    bNegative = isNumeric_Data_Negative(nvar.data)

    if round_var(nvar, dscale):
        return ""

    workData = copy_128(nvar.data)

    if bNegative:
        if negate_128(workData):
            return ""

    for tmp in range(NUMERIC_MAX_PRECISION):
        unbiasedDigits.append(div10_128(workData, workData))

    unbiasedDigits = unbiasedDigits[::-1]

    for tmp in range(NUMERIC_MAX_PRECISION):
        # suppress leading zeros, but force output of a digit before implied decimal point
        if (tmp < NUMERIC_MAX_PRECISION-dscale-1) and bLeadingZeroes and (unbiasedDigits[tmp] == 0):
            continue
        bLeadingZeroes = False
        work.append((unbiasedDigits[tmp] + 0))
        pos += 1

    work.append(0) # terminate sork string

    tmp = pos # strlen of work data

    if bNegative:
        res += '-'

    if dscale != 0:
        iplaces = tmp - dscale #value before decimal
        for i in range(iplaces):
            res += str(work[i])

        res += '.' #decimal point
        pos +=1
        for i in range(dscale):    #1 more size to copy \0
            res += str(work[i+iplaces])

    else:
        for i in range(tmp):
            res += str(work[i])

    dstSpace = res
    return dstSpace

def mul_128(v1, v2, vRes):
    # We treat the arguments as having 8 16-bit digits and do long multiplications
    # as in the days of the 3 Rs
    #
    # Here's an example with 3 digit numbers written ABC and DEF.  C and F are in the
    # units' (i.e. "base to the zeroth power) position, B and E in the "base" position,
    # and A and D in the "base squared" position.
    #
    # The units position of the product will be the low digit of C*F.
    # The "base" position of the product will be the low digit of (B*F+C*E) plus
    # the carry digit from the first step (this is the high digit of the product C*F.
    # The "base squared" position of the product will be the low digit of (A*F+B*E+C*D)
    # plus carry.  The "base cubed" position of the product will be the low digit of
    # (A*E+B*D) plus carry. The "base to the fourth" position of the product will be
    # the low digit of A*D plus carry.  And the "base to the fifth" position of the
    # product will be the carry.
    #
    # We load the 4 32-bit digit value of "v1" into the 8 16-bit digit value "a",
    # load "v1" into "b", compute the 32-bit sums of products, as in the example
    # above, and store them into the 15 32-bit digit "work" value.  Finally, we
    # step through the "work" value entries, adding any carry, and assigning the low
    # 16-bit digit of each entry to the corresponding 16-bit digit of the 16 16-bit
    # digit result "c"
    #
    # !FIX-jpb this should be optimized

    if USE_MUL_DOUBLE:
        lv = hv = 0
        h1 = (v1[0] << 32) + v1[1]
        l1 = (v1[2] << 32) + v1[3]
        h2 = (v2[0] << 32) + v2[1]
        l2 = (v2[2] << 32) + v2[3]
        bRetVal = mul_double(l1, h1, l2, h2, lv, hv)
        vRes[0] = int((hv & HI32_MASK) >> 32)
        vRes[1] = int(hv & 0xffffffff)
        vRes[2] = int((lv & HI32_MASK) >> 32)
        vRes[3] = int(lv & 0xffffffff)
        return bRetVal

    else:
        carry = 0

        if (v1[0] | v1[1] | v1[2] | v2[0] | v2[1] | v2[2]) == 0:
            val = v1[3] * v2[3]
            vRes[3] = int(val & 0xFFFFFFFF)
            vRes[2] = int(val >> 32)
            vRes[1] = 0
            vRes[0] = 0
            return False

        bResNegative = (isNumeric_Data_Negative(v1) != isNumeric_Data_Negative(v2))
        v1abs = copy_128(v1)
        if isNumeric_Data_Negative(v1abs):
            if negate_128(v1abs):
                return True

        v2abs = copy_128(v2)
        if isNumeric_Data_Negative(v2abs):
            if negate_128(v2abs):
                return True

        a = load_8_digit(v1abs)
        b = load_8_digit(v2abs)

        #TODO: w below is not yet defined. Was this code tested?
        w.append(a[0]*b[0])
        w.append(a[1]*b[0] + a[0]*b[1])
        w.append(a[2]*b[0] + a[1]*b[1] + a[0]*b[2])
        w.append(a[3]*b[0] + a[2]*b[1] + a[1]*b[2] + a[0]*b[3])
        w.append(a[4]*b[0] + a[3]*b[1] + a[2]*b[2] + a[1]*b[3] + a[0]*b[4])
        w.append(a[5]*b[0] + a[4]*b[1] + a[3]*b[2] + a[2]*b[3] + a[1]*b[4] + a[0]*b[5])
        w.append(a[6]*b[0] + a[5]*b[1] + a[4]*b[2] + a[3]*b[3] + a[2]*b[4] + a[1]*b[5] + a[0]*b[6])
        w.append(a[7]*b[0] + a[6]*b[1] + a[5]*b[2] + a[4]*b[3] + a[3]*b[4] + a[2]*b[5] + a[1]*b[6] + a[0]*b[7])
        w.append(a[7]*b[1] + a[6]*b[2] + a[5]*b[3] + a[4]*b[4] + a[3]*b[5] + a[2]*b[6] + a[1]*b[7])
        w.append(a[7]*b[2] + a[6]*b[3] + a[5]*b[4] + a[4]*b[5] + a[3]*b[6] + a[2]*b[7])
        w.append(a[7]*b[3] + a[6]*b[4] + a[5]*b[5] + a[4]*b[6] + a[3]*b[7])
        w.append(a[7]*b[4] + a[6]*b[5] + a[5]*b[6] + a[4]*b[7])
        w.append(a[7]*b[5] + a[6]*b[6] + a[5]*b[7])
        w.append(a[7]*b[6] + a[6]*b[7])
        w.append(a[7] * b[7])

        i = 15
        while i > 0:
            w[i-1] += carry
            #TODO: same for c - it is not defined yet.
            c.append(int(w[i-1] & 0xffff))
            carry = (int(w[i-1] >> 16))
            i -=1

        c.append(carry) # hi order digit is final carry
        c = c[::-1] # reverse list
        bRetVal = store_8_digit_from_16(c[:], vRes)

        if bResNegative:
            if negate_128(vRes):
                return True

    return bRetVal

def load_8_digit(src):
    dest = []
    for i in range(MAX_NUMERIC_DIGIT_COUNT):
        dest.append(src[i] >> 16)
        dest.append(src[i] & 0xffffffff)

    return dest

# tests the hi order 8 digits of src[] for overflow and stores the low order 8 in dest
def store_8_digit_from_16(src, dest):
    for i in range(2*MAX_NUMERIC_DIGIT_COUNT):
        if src[i] != 0:
            return True # overflow

    for j in range(MAX_NUMERIC_DIGIT_COUNT):
        dest[j] = int((src[i] << 16) + src[i+1])
        i += 2

    return False

# mul10_and_add multiplies non-negative TNumericData in place by 10 and adds an int
def mul10_and_add(data, adder):
    i = MAX_NUMERIC_DIGIT_COUNT - 1
    carry = adder

    while i >= 0:
        work = data[i]*10 + carry
        data[i] = int(work & 0xffffffff)
        carry = (work >> 32)
        i -= 1

    return (carry != 0) # true=> overflow

def power_of_10(exponent):
    powersOfTen = []
    next = []
    needsInit = True

    if needsInit:
        next.append(0)
        next.append(0)
        next.append(0)
        next.append(1)

        for i in range(NUMERIC_MAX_PRECISION):
            powersOfTen[i].append(next[0])
            powersOfTen[i].append(next[1])
            powersOfTen[i].append(next[2])
            powersOfTen[i].append(next[3])
            if mul10_and_add(next, 0): # use convenient helper routine in this one-time initing
                #assert(false);          # shouldn't happen if our loop limit correct
                pass

        needsInit = False

    if exponent < NUMERIC_MAX_PRECISION:
        return powersOfTen[exponent]
    elif exponent == NUMERIC_MAX_PRECISION:
        return powersOfTen[0]       # NUMERIC_MAX_PRECISIONth needed for get_digit_count, but entry not used
    else:
        return None                 # This will never arise as its made sure scale will limit to MAX_NUMERIC_DIGIT_COUNT

# Negate a doubleword integer with doubleword result.
# Return nonzero if the operation overflows, assuming it's signed.
# The argument is given as two `int64' pieces in L1 and H1.
# The value is stored as two `int64' pieces in *LV and *HV.
def neg_double(l1, h1, lv, hv):
    if l1 == 0:
        lv = 0 #TODO: value isn't used anywhere. Revise or remove.
        hv = -h1
        if (hv & h1) < 0:
            return 1
        else:
            return 0

    else:
        lv = -l1 #TODO: value isn't used anywhere. Revise or remove.
        hv = ~h1 #TODO: value isn't used anywhere. Revise or remove.
        return 0

# Divide doubleword integer LNUM, HNUM by doubleword integer LDEN, HDEN
# for a quotient (stored in *LQUO, *HQUO) and remainder (in *LREM, *HREM).
# CODE is a tree code for a kind of division, one of
# TRUNC_DIV_EXPR, FLOOR_DIV_EXPR, CEIL_DIV_EXPR, ROUND_DIV_EXPR
# or EXACT_DIV_EXPR
# It controls how the quotient is rounded to a integer.
# Return nonzero if the operation overflows.
# UNS nonzero says do unsigned division.
def div_and_round_double(uns, lnum_orig, hnum_orig, lden_orig, hden_orig, lquo, hquo, lrem, hrem):
    num = [0, 0, 0, 0, 0] # extra element for scaling.
    den = [0, 0, 0, 0]
    quo = [0, 0, 0, 0]
    carry = overflow = quo_neg = 0   #register UNSIGNEDINT64 carry = 0;
    lnum = lnum_orig
    hnum = hnum_orig
    lden = lden_orig
    hden = hden_orig

    # calculate quotient sign and convert operands to unsigned.
    if uns != 0:
        if hnum < 0:
            quo_neg = ~quo_neg #~ quo_neg;
            # (minimum integer) / (-1) is the only overflow case.
            if (neg_double(lnum, hnum, lnum, hnum) != 0) and ((lden & hden) == -1):
                overflow = 1
        if hden < 0:
            quo_neg = ~quo_neg  #~ quo_neg;
            neg_double(lden, hden, lden, hden)

    if hnum == 0 and hden == 0: # single precision
        hquo = hrem = 0
        # This unsigned division rounds toward zero.
        lquo = int((lnum) / (lden))
        # if result is negative, make it so.
        if quo_neg != 0:
            neg_double(lquo, hquo, lquo, hquo)
        # compute trial remainder:  rem = num - (quo * den)
        mul_double(lquo, hquo, lden_orig, hden_orig, lrem, hrem)
        neg_double(lrem, hrem, lrem, hrem)
        add_double(lnum_orig, hnum_orig, lrem, hrem, lrem, hrem)

        return overflow

    if hnum == 0: # trivial case: dividend < divisor
        hquo = lquo = 0
        hrem = hnum
        lrem = lnum
        # if result is negative, make it so.
        if quo_neg != 0:
            neg_double(lquo, hquo, lquo, hquo)
        # compute trial remainder:  rem = num - (quo * den)
        mul_double(lquo, hquo, lden_orig, hden_orig, lrem, hrem)
        neg_double(lrem, hrem, lrem, hrem)
        add_double(lnum_orig, hnum_orig, lrem, hrem, lrem, hrem)

        return overflow

    encodeNum(num[:4], lnum, hnum)
    encodeNum(den[:4], lden, hden)

    # Special code for when the divisor < BASE
    if hden == 0 and (lden < base()):
        # hnum != 0 already checked.
        i = 4 - 1
        while i >= 0:
            work = (num[i]) + carry*base()
            quo[i] = int((work) / (lden))
            carry = work % (lden)
            i -=1
    else:
        # Full double precision division,
        # with thanks to Don Knuth's "Seminumerical Algorithms".
        # Find the highest non-zero divisor digit.
        i = 4 - 1
        while(1):
            if den[i] != 0:
                den_hi_sig = i
                break
            i -=1

        # Insure that the first digit of the divisor is at least BASE/2.
        # This is required by the quotient digit estimation algorithm.

        scale = int(base() / (den[den_hi_sig]+1))
        if scale > 1 :  # scale divisor and dividend
            carry = 0
            i = 0
            while i <= 4-1:
                work = (num[i] * scale) + carry
                num[i] = lowPart(work)
                carry = highPart(work)
                i +=1

            num[4] = carry
            carry = 0
            i = 0
            while i <= 4-1:
                work = (den[i] * scale) + carry
                den[i] = lowPart(work)
                carry = highPart(work)
                if den[i] != 0:
                    den_hi_sig = i
                i +=1

        num_hi_sig = 4

        # Main loop
        i = num_hi_sig - den_hi_sig - 1
        while i >= 0:
            # guess the next quotient digit, quo_est, by dividing the first
            # two remaining dividend digits by the high order quotient digit.
            # quo_est is never low and is at most 2 high.
            num_hi_sig = i + den_hi_sig + 1
            work = num[num_hi_sig]*base() + num[num_hi_sig-1]
            if num[num_hi_sig] != den[den_hi_sig]:
                quo_est = int(work / den[den_hi_sig])
            else:
                quo_est = base() - 1

            # refine quo_est so it's usually correct, and at most one high.
            tmp = work - quo_est*den[den_hi_sig]
            if tmp < base() and den[den_hi_sig-1]*quo_est > (tmp*base()+num[num_hi_sig-2]):
                quo_est -=1

            # Try QUO_EST as the quotient digit, by multiplying the
            # divisor by QUO_EST and subtracting from the remaining dividend.
            # Keep in mind that QUO_EST is the I - 1st digit.

            carry = 0
            j = 0
            while j <= den_hi_sig:
                work = quo_est*den[j] + carry
                carry = highPart(work)
                work = num[i+j] - lowPart(work)
                num[i+j] = lowPart(work)
                if highPart(work) != 0:
                    carry = carry + 1
                else:
                    carry = 0
                j +=1

            # if quo_est was high by one, then num[i] went negative and
            # we need to correct things.

            if num[num_hi_sig] < carry:
                quo_est -=1
                carry = 0 # add divisor back in
                j = 0
                while j <= den_hi_sig:
                    work = num[i+j] + den[j] + carry
                    carry = highPart(work)
                    num[i+j] = lowPart(work)
                    j +=1
                num[num_hi_sig] += carry

            # store the quotient digit.
            quo[i] = quo_est
            i -=1

    decodeNum(quo[:4], lquo, hquo)

    # if result is negative, make it so.
    if quo_neg != 0:
        neg_double(*lquo, *hquo, lquo, hquo)

    # compute trial remainder:  rem = num - (quo * den)
    mul_double(*lquo, *hquo, lden_orig, hden_orig, lrem, hrem)
    neg_double(*lrem, *hrem, lrem, hrem)
    add_double(lnum_orig, hnum_orig, *lrem, *hrem, lrem, hrem)

    return overflow



import datetime
import os
import time

import nzpy

import pytest


@pytest.fixture
def has_tzset():

    # Neither Windows nor Jython 2.5.3 have a time.tzset() so skip
    if hasattr(time, 'tzset'):
        os.environ['TZ'] = "UTC"
        time.tzset()
        return True
    return False


# DBAPI compatible interface tests
@pytest.fixture
def db_table(con, has_tzset):
    with con.cursor() as c:
        c.execute(
            "CREATE TABLE t1 "
            "(f1 int primary key, f2 int not null, f3 varchar(50) null) "
            )
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (?, ?, ?)", (1, 1, ''))
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (?, ?, ?)", (2, 10, ''))
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (?, ?, ?)", (3, 100, ''))
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (?, ?, ?)", (4, 1000, ''))
        c.execute(
            "INSERT INTO t1 (f1, f2, f3) VALUES (?, ?, ?)",
            (5, 10000, ''))
    return con


def test_parallel_queries(db_table):
    with db_table.cursor() as c1, db_table.cursor() as c2:

        c1.execute("SELECT f1, f2, f3 FROM t1")
        while 1:
            row = c1.fetchone()
            if row is None:
                break
            f1, f2, f3 = row
            c2.execute("SELECT f1, f2, f3 FROM t1 WHERE f1 > ?", (f1,))
            while 1:
                row = c2.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
        c1.execute("drop table t1")


def test_qmark(db_table):
    orig_paramstyle = nzpy.paramstyle
    try:
        nzpy.paramstyle = "qmark"
        with db_table.cursor() as c1:
            c1.execute("SELECT f1, f2, f3 FROM t1 WHERE f1 > ?", (3,))
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
            c1.execute("drop table t1")
    finally:
        nzpy.paramstyle = orig_paramstyle


def test_arraysize(db_table):
    with db_table.cursor() as c1:
        c1.arraysize = 3
        c1.execute("SELECT * FROM t1")
        retval = c1.fetchmany()
        assert len(retval) == c1.arraysize
        c1.execute("drop table t1")


def test_date():
    val = nzpy.Date(2001, 2, 3)
    assert val == datetime.date(2001, 2, 3)


def test_time():
    val = nzpy.Time(4, 5, 6)
    assert val == datetime.time(4, 5, 6)


def test_timestamp():
    val = nzpy.Timestamp(2001, 2, 3, 4, 5, 6)
    assert val == datetime.datetime(2001, 2, 3, 4, 5, 6)


def test_date_from_ticks(has_tzset):
    if has_tzset:
        val = nzpy.DateFromTicks(1173804319)
        assert val == datetime.date(2007, 3, 13)


def testTimeFromTicks(has_tzset):
    if has_tzset:
        val = nzpy.TimeFromTicks(1173804319)
        assert val == datetime.time(16, 45, 19)


def test_timestamp_from_ticks(has_tzset):
    if has_tzset:
        val = nzpy.TimestampFromTicks(1173804319)
        assert val == datetime.datetime(2007, 3, 13, 16, 45, 19)


def test_binary():
    v = nzpy.Binary(b"\x00\x01\x02\x03\x02\x01\x00")
    assert v == b"\x00\x01\x02\x03\x02\x01\x00"
    assert isinstance(v, nzpy.BINARY)


def test_row_count(db_table):
    with db_table.cursor() as c1:
        c1.execute("SELECT count(*) FROM t1")

        assert [5] == c1.fetchall()[0]

        c1.execute("UPDATE t1 SET f3 = ? WHERE f2 > 101", ("Hello!",))
        assert 2 == c1.rowcount

        c1.execute("DELETE FROM t1")
        assert 5 == c1.rowcount
        c1.execute("drop table t1")

#TODO
@pytest.mark.skip(
        """Skipping. fetchmany() method needs to be fixed."""
    )
def test_fetch_many(db_table):
    with db_table.cursor() as cursor:
        cursor.arraysize = 2
        cursor.execute("SELECT * FROM t1")
        assert 2 == len(cursor.fetchmany())
        assert 2 == len(cursor.fetchmany())
        assert 1 == len(cursor.fetchmany())
        assert 0 == len(cursor.fetchmany())
        cursor.execute("drop table t1")


def test_iterator(db_table):
    with db_table.cursor() as cursor:
        cursor.execute("SELECT * FROM t1 ORDER BY f1")
        f1 = 0
        for row in cursor:
            next_f1 = row[0]
            assert next_f1 > f1
            f1 = next_f1
        cursor.execute("drop table t1")


# Vacuum can't be run inside a transaction, so we need to turn
# autocommit on.
def test_vacuum(con):
    con.autocommit = True
    with con.cursor() as cursor:
        cursor.execute("vacuum")


def test_cursor_type(cursor):
    assert str(type(cursor)) == "<class 'nzpy.core.Cursor'>"


'''
def test_prepared_statement(con):
    with con.cursor() as cursor:
        cursor.execute('PREPARE gen_series AS SELECT generate_series(1, 10);')
        cursor.execute('EXECUTE gen_series')
def test_numeric(db_table):
    orig_paramstyle = nzpy.paramstyle
    try:
        nzpy.paramstyle = "numeric"
        with db_table.cursor() as c1:
            c1.execute("SELECT f1, f2, f3 FROM t1 WHERE f1 > :1", (3,))
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        nzpy.paramstyle = orig_paramstyle
def test_named(db_table):
    orig_paramstyle = nzpy.paramstyle
    try:
        nzpy.paramstyle = "named"
        with db_table.cursor() as c1:
            c1.execute(
                "SELECT f1, f2, f3 FROM t1 WHERE f1 > :f1", {"f1": 3})
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        nzpy.paramstyle = orig_paramstyle
def test_format(db_table):
    orig_paramstyle = nzpy.paramstyle
    try:
        nzpy.paramstyle = "format"
        with db_table.cursor() as c1:
            c1.execute("SELECT f1, f2, f3 FROM t1 WHERE f1 > %s", (3,))
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        nzpy.paramstyle = orig_paramstyle
def test_pyformat(db_table):
    orig_paramstyle = nzpy.paramstyle
    try:
        nzpy.paramstyle = "pyformat"
        with db_table.cursor() as c1:
            c1.execute(
                "SELECT f1, f2, f3 FROM t1 WHERE f1 > %(f1)s", {"f1": 3})
            while 1:
                row = c1.fetchone()
                if row is None:
                    break
                f1, f2, f3 = row
    finally:
        nzpy.paramstyle = orig_paramstyle
'''



import io
import logging
import pathlib

import nzpy

import pytest


@pytest.fixture
def lf(request):
    f = pathlib.Path("nzpy.log")
    if f.exists():
        f.unlink()
    try:
        yield f
    finally:
        if f.exists():
            f.unlink()


@pytest.fixture
def db(db_kwargs):
    db_kwargs['database'] = 'system'
    return db_kwargs


@pytest.fixture
def base_log(request):
    for log in logging.getLogger().handlers.copy():
        logging.getLogger().removeHandler(log)
    lbuf = io.StringIO()

    s = logging.StreamHandler(lbuf)
    s.setFormatter(logging.Formatter('%(name)s %(message)s'))
    logging.getLogger().addHandler(s)
    return lbuf


def test_parent_logging(db, lf, base_log):
    db['logOptions'] = nzpy.LogOptions.Inherit
    nzpy.connect(**db)

    # log file shuoldn't be created
    assert not lf.exists()
    # parent should have its logs
    assert 'nzpy' in base_log.getvalue()


def test_parent_logging_negative(db, lf, base_log):
    db['logLevel'] = logging.INFO

    for i in [logging.getLogger()] + logging.getLogger().handlers:
        i.setLevel(logging.CRITICAL)

    logging.critical("hello")
    nzpy.connect(**db)
    # log file shuoldn't be created
    assert not lf.exists()
    # parent shouldn't have its logs
    assert 'nzpy' not in base_log.getvalue()


def test_parent_logging_disabled(db, lf, base_log):
    # db['log_option'] = ~(nzpy.LogOptions.Inherit|nzpy.LogOptions.Logfile)
    db['logOptions'] = nzpy.LogOptions.Disabled
    db['logLevel'] = logging.INFO
    logging.basicConfig()
    logging.error("hello")
    nzpy.connect(**db)
    # log file shuoldn't be created
    assert not lf.exists()
    # parent shouldn't have its logs
    assert 'nzpy' not in base_log.getvalue()


def test_log_file(db, lf, base_log):
    db['logOptions'] = nzpy.LogOptions.Logfile
    nzpy.connect(**db)

    assert lf.exists()
    # parent should have its logs
    assert 'nzpy' not in base_log.getvalue()
    nzpy.connect(**db)



#!/usr/bin/env python
import os

from setuptools import setup

readme = os.path.join(os.path.dirname(__file__), "README.md")

setup(
    name="nzpy",
    version='1.17',
    description="IBM Netezza python driver",
    long_description=open(readme).read(),
    long_description_content_type='text/markdown',
    author="IBM",
    author_email="shabmoh3@in.ibm.com",
    url="https://github.com/ibm/nzpy",
    license="IBM",
    python_requires='>=3.5',
    install_requires=['scramp==1.4.5', 'looseversion==1.3.0'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: Jython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
        "Topic :: Database :: Front-Ends",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        "Documentation": "https://github.com/IBM/nzpy/wiki",
        "Source": "https://github.com/IBM/nzpy",
        "Tracker": "https://github.com/IBM/nzpy/issues",
    },

    keywords="Netezza dbapi",
    packages=("nzpy",)
)



import base64
import ssl
from getpass import getuser
from hashlib import md5
from hashlib import sha256
from os import getpid, path
from platform import system
from socket import gethostname
from sys import argv

from nzpy import core

#  CP Version
CP_VERSION_1 = 1
CP_VERSION_2 = 2
CP_VERSION_3 = 3
CP_VERSION_4 = 4
CP_VERSION_5 = 5
CP_VERSION_6 = 6

#  Handshake version
HSV2_INVALID_OPCODE = 0
HSV2_CLIENT_BEGIN = 1
HSV2_DB = 2
HSV2_USER = 3
HSV2_OPTIONS = 4
HSV2_TTY = 5
HSV2_REMOTE_PID = 6
HSV2_PRIOR_PID = 7
HSV2_CLIENT_TYPE = 8
HSV2_PROTOCOL = 9
HSV2_HOSTCASE = 10
HSV2_SSL_NEGOTIATE = 11
HSV2_SSL_CONNECT = 12
HSV2_APPNAME = 13
HSV2_CLIENT_OS = 14
HSV2_CLIENT_HOST_NAME = 15
HSV2_CLIENT_OS_USER = 16
HSV2_64BIT_VARLENA_ENABLED = 17
HSV2_CLIENT_DONE = 1000

#  PG PROTOCOL
PG_PROTOCOL_3 = 3
PG_PROTOCOL_4 = 4
PG_PROTOCOL_5 = 5

#  Authentication types
AUTH_REQ_OK = 0
AUTH_REQ_KRB4 = 1
AUTH_REQ_KRB5 = 2
AUTH_REQ_PASSWORD = 3
AUTH_REQ_CRYPT = 4
AUTH_REQ_MD5 = 5
AUTH_REQ_SHA256 = 6

#  Client type
NPS_CLIENT = 0
IPS_CLIENT = 1

#  Client type python
NPSCLIENT_TYPE_PYTHON = 13


class Handshake():

    def __init__(self, _usock, _sock, ssl, log):
        self._hsVersion = None
        self._protocol1 = None
        self._protocol2 = None
        self._usock = _usock
        self._sock = _sock
        self.ssl_params = ssl
        self.log = log

        #  guardium related information
        self.guardium_clientOS = system()
        self.guardium_clientOSUser = getuser()
        self.guardium_clientHostName = gethostname()
        self.guardium_applName = path.basename(argv[0])

    def startup(self, database, securityLevel, user, password, pgOptions):
        #  Negotiate the handshake version (connection protocol)
        if not self.conn_handshake_negotiate(self._sock.write, self._sock.read,
                                             self._sock.flush, self._hsVersion,
                                             self._protocol2):
            self.log.info("Handshake negotiation unsuccessful")
            return False

        self.log.debug("Sending handshake information to server")
        if not self.conn_send_handshake_info(self._sock.write, self._sock.read,
                                             self._sock.flush, database,
                                             securityLevel, self._hsVersion,
                                             self._protocol1, self._protocol2,
                                             user, pgOptions):
            self.log.warning("Error in conn_send_handshake_info")
            return False

        if not self.conn_authenticate(self._sock.write, self._sock.read,
                                      self._sock.flush, password):
            self.log.warning("Error in conn_authenticate")
            return False

        if not self.conn_connection_complete(self._sock.read):
            self.log.warning("Error in conn_connection_complete")
            return False

        return self._sock

    def conn_handshake_negotiate(self, _write, _read, _flush,
                                 _hsVersion, _protocol2):
        version = CP_VERSION_6
        self.log.debug("Latest-handshake version (conn-protocol) "
                       "= %s", version)

        while(1):
            if version == CP_VERSION_6:
                version = CP_VERSION_6

            if version == CP_VERSION_5:
                version = CP_VERSION_5

            if version == CP_VERSION_4:
                version = CP_VERSION_4

            if version == CP_VERSION_3:
                version = CP_VERSION_3

            if version == CP_VERSION_2:
                version = CP_VERSION_2

            self.log.debug("sending version: %s", version)
            val = bytearray(core.h_pack(HSV2_CLIENT_BEGIN) +
                            core.h_pack(version))
            _write(core.i_pack(len(val) + 4))
            _write(val)
            _flush()

            self.log.info("sent handshake negotiation block successfully")

            beresp = _read(1)

            self.log.debug("Got response: %s", beresp)

            if beresp == b'N':
                self._hsVersion = version
                self._protocol2 = 0
                return True
            elif beresp == b'M':
                version = _read(1)
                if version == b'2':
                    version = CP_VERSION_2
                if version == b'3':
                    version = CP_VERSION_3
                if version == b'4':
                    version = CP_VERSION_4
                if version == b'5':
                    version = CP_VERSION_5
            elif beresp == b"E":
                self.log.warning("Bad attribute value error")
                return False
            else:
                self.log.warning("Bad protocol error")
                return False

    def conn_send_handshake_info(self, _write, _read, _flush, _database,
                                 securityLevel, _hsVersion,
                                 _protocol1, _protocol2,
                                 user, pgOptions):
        #  We need database information at the backend in order to
        #  select security restrictions. So always send the database first
        if not self.conn_send_database(_write, _read, _flush, _database):
            return False

        #  If the backend supports security features and if the driver
        #  requires secured session, negotiate security requirements now
        if not self.conn_secure_session(securityLevel):
            return False

        if not self.conn_set_next_dataprotocol(self._protocol1,
                                               self._protocol2):
            return False

        if self._hsVersion == CP_VERSION_6 or \
                self._hsVersion == CP_VERSION_4:
            return self.conn_send_handshake_version4(self._sock.write,
                                                     self._sock.read,
                                                     self._sock.flush,
                                                     self._protocol1,
                                                     self._protocol2,
                                                     self._hsVersion,
                                                     user, pgOptions)
        elif self._hsVersion == CP_VERSION_5 or self._hsVersion == \
                CP_VERSION_3 or self._hsVersion == CP_VERSION_2:
            return self.conn_send_handshake_version2(self._sock.write,
                                                     self._sock.read,
                                                     self._sock.flush,
                                                     self._protocol1,
                                                     self._protocol2,
                                                     self._hsVersion,
                                                     user, pgOptions)

        return True

    def conn_send_database(self, _write, _read, _flush, _database):
        db = None
        if _database is not None:
            if isinstance(_database, str):
                db = _database.encode('utf8')
            self.log.info("Database name: %s", str(db, 'utf8'))

            val = bytearray(core.h_pack(HSV2_DB))
            val.extend(db + core.NULL_BYTE)
            _write(core.i_pack(len(val) + 4))
            _write(val)
            _flush()

        beresp = _read(1)
        self.log.info("Backend response: %s", str(beresp, 'utf8'))
        if beresp == b'N':
            return True
        elif beresp == core.ERROR_RESPONSE:
            self.log.warning("ERROR_AUTHOR_BAD")
            return False
        else:
            self.log.warning("Unknown response")
            return False

    def conn_set_next_dataprotocol(self, _protocol1, _protocol2):
        if self._protocol2 == 0:
            self._protocol2 = PG_PROTOCOL_5
        elif _protocol2 == 5:
            self._protocol2 = PG_PROTOCOL_4
        elif _protocol2 == 4:
            self._protocol2 = PG_PROTOCOL_3
        else:
            return False

        self._protocol1 = PG_PROTOCOL_3
        self.log.debug("Connection protocol set to : %s %s",
                       self._protocol1, self._protocol2)
        return True

    def conn_secure_session(self, securityLevel):
        information = HSV2_SSL_NEGOTIATE
        currSecLevel = securityLevel

        while information != 0:
            if information == HSV2_SSL_NEGOTIATE:
                # SecurityLevel meaning
                # ---------------------------------------
                #      0    Preferred Unsecured session
                #      1    Only Unsecured session
                #      2    Preferred Secured session
                #      3    Only Secured session
                #
                self.log.debug("Security Level requested = %s", currSecLevel)
                opcode = information

            if information == HSV2_SSL_CONNECT:
                opcode = information

            val = bytearray(core.h_pack(opcode) + core.i_pack(currSecLevel))
            self._sock.write(core.i_pack(len(val) + 4))
            self._sock.write(val)
            self._sock.flush()

            if information == HSV2_SSL_CONNECT:
                try:
                    self._usock = ssl.SSLContext().wrap_socket(self._usock)
                    self._sock = self._usock.makefile(mode="rwb")
                    self.log.info("Secured Connect Success")
                except ssl.SSLError:
                    self.log.warning("Problem establishing "
                                     "secured session")
                    return False

            if information != 0:
                beresp = self._sock.read(1)
                self.log.debug("Got response =%s", beresp)
                if beresp == b'S':
                    #  The backend sends 'S' only in 3 cases
                    #  Client requests strict SSL and backend
                    #  supports it. Client requests preferred
                    #  SSL and backend supports it. Client
                    #  requests preferred non-SSL, but
                    #  backend supports only secured sessions.

                    if not isinstance(self.ssl_params, dict):
                        self.ssl_params = {}
                    try:
                        import ssl

                        ca_certs = self.ssl_params.get('ca_certs')
                        ssl_context = ssl.create_default_context(
                            cafile=ca_certs)
                        ssl_context.check_hostname = False
                        if ca_certs is None:
                            ssl_context.verify_mode = ssl.CERT_NONE
                        else:
                            ssl_context.verify_mode = ssl.CERT_REQUIRED

                        information = HSV2_SSL_CONNECT

                    except ImportError:
                        raise core.InterfaceError("SSL required but ssl "
                                             "module not available in this "
                                             "python installation")

                    except ssl.SSLError:
                        if currSecLevel == 2:
                            self.log.debug("Problem establishing "
                                           "secured session")
                            self.log.debug("Attempting unsecured session")
                            currSecLevel = 1
                            information = HSV2_SSL_NEGOTIATE
                            continue
                        self.log.warning("Problem establishing "
                                         "secured session")
                        return False

                if beresp == b'N':
                    if information == HSV2_SSL_NEGOTIATE:
                        self.log.debug("Attempting unsecured session")
                    information = 0
                    return True

                if beresp == b'E':
                    #  If 'E' is received, because the SSL
                    #  session establishment failed, and we are
                    #  requesting preferred secured session,
                    #  we will have to attempt a non secured session.
                    #  If this also fails, we have to error out.
                    #  To achieve this, we now negotiate for
                    #  essential non-secured session
                    self.log.warning("Error: connection failed")
                    return False

    def conn_send_handshake_version2(self, _write, _read, _flush, _protocol1,
                                     _protocol2, _hsVersion, user, pgOptions):
        if isinstance(user, str):
            user = user.encode('utf8')

        information = HSV2_USER
        val = bytearray(core.h_pack(information))
        val.extend(user + core.NULL_BYTE)
        information = HSV2_PROTOCOL

        while information != 0:
            _write(core.i_pack(len(val) + 4))
            _write(val)
            _flush()
            beresp = _read(1)
            self.log.info("Backend response: %s", str(beresp, 'utf8'))
            if beresp == b'N':
                if information == HSV2_PROTOCOL:
                    val = bytearray(core.h_pack(information) +
                                    core.h_pack(_protocol1) +
                                    core.h_pack(_protocol2))
                    information = HSV2_REMOTE_PID
                    continue

                if information == HSV2_REMOTE_PID:
                    val = bytearray(core.h_pack(information) +
                                    core.i_pack(getpid()))
                    information = HSV2_OPTIONS
                    continue

                if information == HSV2_OPTIONS:
                    if pgOptions is not None:
                        val = bytearray(core.h_pack(information))
                        val.extend(pgOptions.encode('utf8') +
                                   core.NULL_BYTE)
                    information = HSV2_CLIENT_TYPE
                    continue

                if information == HSV2_CLIENT_TYPE:
                    val = bytearray(core.h_pack(information) +
                                    core.h_pack(NPSCLIENT_TYPE_PYTHON))
                    if _hsVersion == CP_VERSION_5 or _hsVersion \
                            == CP_VERSION_6:
                        information = HSV2_64BIT_VARLENA_ENABLED
                    else:
                        information = HSV2_CLIENT_DONE
                    continue

                if information == HSV2_64BIT_VARLENA_ENABLED:
                    val = bytearray(core.h_pack(information) +
                                    core.h_pack(IPS_CLIENT))
                    information = HSV2_CLIENT_DONE
                    continue

                if information == HSV2_CLIENT_DONE:
                    val = bytearray(core.h_pack(information))
                    information = 0
                    _write(core.i_pack(len(val) + 4))
                    _write(val)
                    _flush()
                    return True

            elif beresp == core.ERROR_RESPONSE:
                self.log.warning("ERROR_CONN_FAIL")
                return False

    def conn_send_handshake_version4(self, _write, _read, _flush,
                                     _protocol1, _protocol2,
                                     _hsVersion, user, pgOptions):
        if isinstance(user, str):
            user = user.encode('utf8')

        information = HSV2_USER
        val = bytearray(core.h_pack(information))
        val.extend(user + core.NULL_BYTE)
        information = HSV2_APPNAME

        while information != 0:
            _write(core.i_pack(len(val) + 4))
            _write(val)
            _flush()
            beresp = _read(1)
            self.log.info("Backend response: %s", str(beresp, 'utf8'))
            if beresp == b'N':
                if information == HSV2_APPNAME:
                    #  App name
                    val = bytearray(core.h_pack(information))
                    val.extend(self.guardium_applName.encode('utf8') +
                               core.NULL_BYTE)
                    self.log.debug("Appname :%s", self.guardium_applName)
                    information = HSV2_CLIENT_OS
                    continue

                if information == HSV2_CLIENT_OS:
                    #  OS name
                    val = bytearray(core.h_pack(information))
                    val.extend(self.guardium_clientOS.encode('utf8') +
                               core.NULL_BYTE)
                    self.log.debug("Client OS :%s", self.guardium_clientOS)
                    information = HSV2_CLIENT_HOST_NAME
                    continue

                if information == HSV2_CLIENT_HOST_NAME:
                    #  Client Host name
                    val = bytearray(core.h_pack(information))
                    val.extend(self.guardium_clientHostName.encode('utf8') +
                               core.NULL_BYTE)
                    self.log.debug("Client hostname :%s",
                                   self.guardium_clientHostName)
                    information = HSV2_CLIENT_OS_USER
                    continue

                if information == HSV2_CLIENT_OS_USER:
                    #  client OS User name
                    val = bytearray(core.h_pack(information))
                    val.extend(self.guardium_clientOSUser.encode('utf8') +
                               core.NULL_BYTE)
                    self.log.debug("Client OS user :%s",
                                   self.guardium_clientOSUser)
                    information = HSV2_PROTOCOL
                    continue

                if information == HSV2_PROTOCOL:
                    val = bytearray(core.h_pack(information) +
                                    core.h_pack(_protocol1) +
                                    core.h_pack(_protocol2))
                    information = HSV2_REMOTE_PID
                    continue

                if information == HSV2_REMOTE_PID:
                    val = bytearray(core.h_pack(information) +
                                    core.i_pack(getpid()))
                    information = HSV2_OPTIONS
                    continue

                if information == HSV2_OPTIONS:
                    if pgOptions is not None:
                        val = bytearray(core.h_pack(information))
                        val.extend(pgOptions.encode('utf8') + core.NULL_BYTE)
                    information = HSV2_CLIENT_TYPE
                    continue

                if information == HSV2_CLIENT_TYPE:
                    val = bytearray(core.h_pack(information) +
                                    core.h_pack(NPSCLIENT_TYPE_PYTHON))
                    if _hsVersion == CP_VERSION_5 or \
                            _hsVersion == CP_VERSION_6:
                        information = HSV2_64BIT_VARLENA_ENABLED
                    else:
                        information = HSV2_CLIENT_DONE
                    continue

                if information == HSV2_64BIT_VARLENA_ENABLED:
                    val = bytearray(core.h_pack(information) +
                                    core.h_pack(IPS_CLIENT))
                    information = HSV2_CLIENT_DONE
                    continue

                if information == HSV2_CLIENT_DONE:
                    val = bytearray(core.h_pack(information))
                    information = 0
                    _write(core.i_pack(len(val) + 4))
                    _write(val)
                    _flush()
                    return True

            elif beresp == core.ERROR_RESPONSE:
                self.log.warning("ERROR_CONN_FAIL")
                return False

    def conn_authenticate(self, _write, _read, _flush, password):
        if isinstance(password, str):
            password = password.encode('utf8')

        beresp = _read(1)
        self.log.debug("Got response: %s", beresp)

        if beresp != core.AUTHENTICATION_REQUEST:
            self.log.warning("Authentication error")
            return False

        self.log.debug("auth got 'R' - request for password")
        areq = core.i_unpack(_read(4))[0]
        self.log.debug("areq =%s", areq)

        if areq == AUTH_REQ_OK:
            self.log.info("success")
            return True

        if areq == AUTH_REQ_PASSWORD:
            self.log.info("Plain password requested")
            _write(core.i_pack(len(password + core.NULL_BYTE) + 4))
            _write(password + core.NULL_BYTE)
            _flush()

        if areq == AUTH_REQ_MD5:
            self.log.info("Password type is MD5")
            salt = _read(2)
            self.log.debug("Salt =%s", salt)
            if password is None:
                raise core.InterfaceError(
                    "server requesting MD5 password authentication, "
                    "but no password was provided")
            md5encoded = md5(salt+password)
            md5pwd = base64.standard_b64encode(md5encoded.digest())
            pwd = md5pwd.rstrip(b"=")
            self.log.debug("md5 encrypted password is =%s", pwd)

            # Int32 - Message length including self.
            # String - The password.  Password may be encrypted.
            _write(core.i_pack(len(pwd + core.NULL_BYTE) + 4))
            _write(pwd + core.NULL_BYTE)
            _flush()

        if areq == AUTH_REQ_SHA256:
            self.log.info("Password type is SSH")
            salt = _read(2)
            self.log.debug("Salt =%s", salt)
            if password is None:
                raise core.InterfaceError("server requesting MD5 password "
                                     "authentication, but no password "
                                     "was provided")
            sha256encoded = sha256(salt+password)
            sha256pwd = base64.standard_b64encode(sha256encoded.digest())
            pwd = sha256pwd.rstrip(b"=")
            self.log.debug("sha256 encrypted password is =%s", pwd)

            # Int32 - Message length including
            # String - The password.  Password may be encrypted.
            _write(core.i_pack(len(pwd + core.NULL_BYTE) + 4))
            _write(pwd + core.NULL_BYTE)
            _flush()

        if areq == AUTH_REQ_KRB5:
            self.log.info("krb encryption requested from backend")

        return True

    def conn_connection_complete(self, _read):
        while (1):
            response = _read(1)
            self.log.info("backend response: %s", str(response, 'utf8'))

            if response != core.AUTHENTICATION_REQUEST:
                _read(4)
                #  do not use just ignore
                length = core.i_unpack(_read(4))[0]

            if response == core.AUTHENTICATION_REQUEST:
                areq = core.i_unpack(_read(4))[0]
                self.log.info("backend response: %s", areq)

            if response == core.NOTICE_RESPONSE:
                notices = str(_read(length), 'utf8')
                self.log.debug("Response received from "
                               "backend: %s", notices)

            if response == core.BACKEND_KEY_DATA:
                areq = core.i_unpack(_read(4))[0]
                self.log.debug("Backend response PID: %s", areq)

                areq = core.i_unpack(_read(4))[0]
                self.log.debug("Backend response KEY: %s", areq)

            if response == core.READY_FOR_QUERY:
                self.log.info("Authentication Successful")
                return True

            if response == core.ERROR_RESPONSE:
                error = str(_read(length), 'utf8')
                self.log.warning("Error occured, server response:%s", error)
                return False
